from unittest import mock
import unittest
import os
import shutil
import zipfile
import json
from datetime import datetime, timezone
from base64 import b64decode, b64encode

from webscrapbook import WSB_DIR
from webscrapbook import util
from webscrapbook.scrapbook import exporter as wsb_exporter
from webscrapbook._compat import zip_stream

root_dir = os.path.abspath(os.path.dirname(__file__))
test_root = os.path.join(root_dir, 'test_scrapbook_exporter')

def setUpModule():
    # mock out user config
    global mockings
    mockings = [
        mock.patch('webscrapbook.scrapbook.host.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_CONFIG', test_root),
        ]
    for mocking in mockings:
        mocking.start()

def tearDownModule():
    # stop mock
    for mocking in mockings:
        mocking.stop()

class TestExporter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = 8192
        cls.test_root = os.path.join(test_root, 'general')
        cls.test_input = os.path.join(cls.test_root, 'input')
        cls.test_input_wsb = os.path.join(cls.test_input, WSB_DIR)
        cls.test_input_config = os.path.join(cls.test_input_wsb, 'config.ini')
        cls.test_input_tree = os.path.join(cls.test_input_wsb, 'tree')
        cls.test_input_meta = os.path.join(cls.test_input_tree, 'meta.js')
        cls.test_input_toc = os.path.join(cls.test_input_tree, 'toc.js')
        cls.test_output = os.path.join(cls.test_root, 'output')

    def setUp(self):
        """Set up a general temp test folder
        """
        os.makedirs(self.test_input_tree, exist_ok=True)
        os.makedirs(self.test_output, exist_ok=True)

    def tearDown(self):
        """Remove general temp test folder
        """
        try:
            shutil.rmtree(self.test_root)
        except NotADirectoryError:
            os.remove(self.test_root)
        except FileNotFoundError:
            pass

    def test_basic01(self):
        """Test exporting a common */index.html
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder",
    "title": "item0",
    "index": "20200101000000000/index.html",
    "create": "20200102000000000",
    "modify": "20200103000000000",
    "source": "http://example.com",
    "icon": "favicon.bmp"
  }
})""")
        with open(self.test_input_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")
        index_file = os.path.join(self.test_input, '20200101000000000', 'index.html')
        os.makedirs(os.path.dirname(index_file))
        with open(index_file, 'w', encoding='UTF-8') as fh:
            fh.write('ABC123')

        for info in wsb_exporter.run(self.test_input, self.test_output):
            pass

        with os.scandir(self.test_output) as entries:
            files = sorted(entries, key=lambda x: x.path)

        # files are exported in depth-first order
        with zipfile.ZipFile(files[0]) as zh:
            with zh.open('meta.json') as fh:
                data = json.load(fh)
            with zh.open('export.json') as fh:
                export_info = json.load(fh)
            with zh.open('data/20200101000000000/index.html') as fh:
                index_data = fh.read().decode('UTF-8')
                
        self.assertEqual(data, {
            'id': '20200101000000000',
            'type': 'folder',
            'title': 'item0',
            'index': '20200101000000000/index.html',
            'create': '20200102000000000',
            'modify': '20200103000000000',
            'source': 'http://example.com',
            'icon': 'favicon.bmp',
            })

        self.assertEqual(export_info['version'], 1)
        self.assertAlmostEqual(util.id_to_datetime(export_info['id']).timestamp(), datetime.now(timezone.utc).timestamp(), delta=3)
        self.assertEqual(export_info['timestamp'], export_info['id'])
        self.assertEqual(export_info['timezone'], datetime.now().astimezone().utcoffset().total_seconds())
        self.assertEqual(export_info['path'], [{'id': 'root', 'title': ''}])

        self.assertEqual(index_data, 'ABC123')

    def test_basic02(self):
        """Test exporting a common *.htz
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder",
    "title": "item0",
    "index": "20200101000000000.htz",
    "create": "20200102000000000",
    "modify": "20200103000000000",
    "source": "http://example.com",
    "icon": ".wsb/tree/favicon/dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp"
  }
})""")
        with open(self.test_input_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ]
})""")
        index_file = os.path.join(self.test_input, '20200101000000000.htz')
        with zipfile.ZipFile(index_file, 'w') as zh:
            zh.writestr('index.html', 'ABC123')
        favicon_file = os.path.join(self.test_input_tree, 'favicon', 'dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp')
        os.makedirs(os.path.dirname(favicon_file))
        with open(favicon_file, 'wb') as fh:
            fh.write(b64decode('Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA'))

        for info in wsb_exporter.run(self.test_input, self.test_output):
            pass

        with os.scandir(self.test_output) as entries:
            files = sorted(entries, key=lambda x: x.path)

        # files are exported in depth-first order
        with zipfile.ZipFile(files[0]) as zh:
            with zh.open('meta.json') as fh:
                data = json.load(fh)
            with zh.open('export.json') as fh:
                export_info = json.load(fh)
            with zh.open('data/20200101000000000.htz') as fh:
                fh = zip_stream(fh)
                with zipfile.ZipFile(fh) as zh2:
                    with zh2.open('index.html') as fh2:
                        index_data = fh2.read().decode('UTF-8')
            with zh.open('favicon/dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp') as fh:
                favicon_data = fh.read()

        self.assertEqual(data, {
            'id': '20200101000000000',
            'type': 'folder',
            'title': 'item0',
            'index': '20200101000000000.htz',
            'create': '20200102000000000',
            'modify': '20200103000000000',
            'source': 'http://example.com',
            'icon': '.wsb/tree/favicon/dbc82be549e49d6db9a5719086722a4f1c5079cd.bmp',
            })

        self.assertEqual(export_info['version'], 1)
        self.assertAlmostEqual(util.id_to_datetime(export_info['id']).timestamp(), datetime.now(timezone.utc).timestamp(), delta=3)
        self.assertEqual(export_info['timestamp'], export_info['id'])
        self.assertEqual(export_info['timezone'], datetime.now().astimezone().utcoffset().total_seconds())
        self.assertEqual(export_info['path'], [{'id': 'root', 'title': ''}])

        self.assertEqual(index_data, 'ABC123')
        self.assertEqual(b64encode(favicon_data), b'Qk08AAAAAAAAADYAAAAoAAAAAQAAAAEAAAABACAAAAAAAAYAAAASCwAAEgsAAAAAAAAAAAAAAP8AAAAA')

    def test_toc01(self):
        """Export all if item_ids not set

        - Include hidden (at last).
        - Exclude recycle.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder",
    "title": "item0"
  },
  "20200101000000001": {
    "type": "folder",
    "title": "item1"
  },
  "20200101000000002": {
    "type": "folder",
    "title": "item2"
  },
  "20200101000000003": {
    "type": "folder",
    "title": "item3"
  },
  "20200101000000004": {
    "type": "folder",
    "title": "item4"
  }
})""")
        with open(self.test_input_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "hidden": [
    "20200101000000003"
  ],
  "root": [
    "20200101000000000",
    "20200101000000001"
  ],
  "20200101000000000": [
    "20200101000000002"
  ],
  "recycle": [
    "20200101000000004"
  ]
})""")

        for info in wsb_exporter.run(self.test_input, self.test_output):
            pass

        with os.scandir(self.test_output) as entries:
            files = sorted(entries, key=lambda x: x.path)
        metas = []
        export_infos = []
        for file in files:
            with zipfile.ZipFile(file) as zh:
                with zh.open('meta.json') as fh:
                    metas.append(json.load(fh))
                with zh.open('export.json') as fh:
                    export_infos.append(json.load(fh))

        self.assertEqual(len(files), 4)
        self.assertEqual(len(set(e['id'] for e in export_infos)), 4)

        # files are exported in depth-first order
        self.assertEqual(metas[0], {
            'id': '20200101000000000',
            'type': 'folder',
            'title': 'item0',
            })
        self.assertEqual(export_infos[0]['path'], [
            {'id': 'root', 'title': ''},
            ])

        self.assertEqual(metas[1], {
            'id': '20200101000000002',
            'type': 'folder',
            'title': 'item2',
            })
        self.assertEqual(export_infos[1]['path'], [
            {'id': 'root', 'title': ''},
            {'id': '20200101000000000', 'title': 'item0'},
            ])

        self.assertEqual(metas[2], {
            'id': '20200101000000001',
            'type': 'folder',
            'title': 'item1',
            })
        self.assertEqual(export_infos[2]['path'], [
            {'id': 'root', 'title': ''},
            ])

        self.assertEqual(metas[3], {
            'id': '20200101000000003',
            'type': 'folder',
            'title': 'item3',
            })
        self.assertEqual(export_infos[3]['path'], [
            {'id': 'hidden', 'title': ''},
            ])

    def test_toc02(self):
        """Export only those specified by item_ids

        - Never include recycle.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder",
    "title": "item0"
  },
  "20200101000000001": {
    "type": "folder",
    "title": "item1"
  },
  "20200101000000002": {
    "type": "folder",
    "title": "item2"
  },
  "20200101000000003": {
    "type": "folder",
    "title": "item3"
  },
  "20200101000000004": {
    "type": "folder",
    "title": "item4"
  },
  "20200101000000005": {
    "type": "folder",
    "title": "item5"
  },
  "20200101000000006": {
    "type": "folder",
    "title": "item6"
  }
})""")
        with open(self.test_input_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "hidden": [
    "20200101000000003",
    "20200101000000004"
  ],
  "root": [
    "20200101000000000",
    "20200101000000001"
  ],
  "20200101000000000": [
    "20200101000000002"
  ],
  "recycle": [
    "20200101000000005",
    "20200101000000006"
  ]
})""")

        for info in wsb_exporter.run(self.test_input, self.test_output,
                item_ids=['20200101000000000', '20200101000000003', '20200101000000005']):
            pass

        with os.scandir(self.test_output) as entries:
            files = sorted(entries, key=lambda x: x.path)
        metas = []
        export_infos = []
        for file in files:
            with zipfile.ZipFile(file) as zh:
                with zh.open('meta.json') as fh:
                    metas.append(json.load(fh))
                with zh.open('export.json') as fh:
                    export_infos.append(json.load(fh))

        self.assertEqual(len(files), 2)

        # files are exported in depth-first order
        self.assertEqual(metas[0], {
            'id': '20200101000000000',
            'type': 'folder',
            'title': 'item0',
            })
        self.assertEqual(export_infos[0]['path'], [
            {'id': 'root', 'title': ''},
            ])

        self.assertEqual(metas[1], {
            'id': '20200101000000003',
            'type': 'folder',
            'title': 'item3',
            })
        self.assertEqual(export_infos[1]['path'], [
            {'id': 'hidden', 'title': ''},
            ])

    def test_toc03(self):
        """Export descendants if recursive"""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder",
    "title": "item0"
  },
  "20200101000000001": {
    "type": "folder",
    "title": "item1"
  },
  "20200101000000002": {
    "type": "folder",
    "title": "item2"
  },
  "20200101000000003": {
    "type": "folder",
    "title": "item3"
  }
})""")
        with open(self.test_input_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000",
    "20200101000000001"
  ],
  "20200101000000000": [
    "20200101000000002"
  ],
  "20200101000000002": [
    "20200101000000003"
  ]
})""")

        for info in wsb_exporter.run(self.test_input, self.test_output,
                item_ids=['20200101000000000'], recursive=True):
            pass

        with os.scandir(self.test_output) as entries:
            files = sorted(entries, key=lambda x: x.path)
        metas = []
        export_infos = []
        for file in files:
            with zipfile.ZipFile(file) as zh:
                with zh.open('meta.json') as fh:
                    metas.append(json.load(fh))
                with zh.open('export.json') as fh:
                    export_infos.append(json.load(fh))

        self.assertEqual(len(files), 3)

        # files are exported in depth-first order
        self.assertEqual(metas[0], {
            'id': '20200101000000000',
            'type': 'folder',
            'title': 'item0',
            })
        self.assertEqual(export_infos[0]['path'], [
            {'id': 'root', 'title': ''},
            ])

        self.assertEqual(metas[1], {
            'id': '20200101000000002',
            'type': 'folder',
            'title': 'item2',
            })
        self.assertEqual(export_infos[1]['path'], [
            {'id': 'root', 'title': ''},
            {'id': '20200101000000000', 'title': 'item0'},
            ])

        self.assertEqual(metas[2], {
            'id': '20200101000000003',
            'type': 'folder',
            'title': 'item3',
            })
        self.assertEqual(export_infos[2]['path'], [
            {'id': 'root', 'title': ''},
            {'id': '20200101000000000', 'title': 'item0'},
            {'id': '20200101000000002', 'title': 'item2'},
            ])

    def test_toc04(self):
        """Export all occurrences

        - Occurrences of the same item should share same export id.
        """
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder",
    "title": "item0"
  },
  "20200101000000001": {
    "type": "folder",
    "title": "item1"
  },
  "20200101000000002": {
    "type": "folder",
    "title": "item2"
  },
  "20200101000000003": {
    "type": "folder",
    "title": "item3"
  }
})""")
        with open(self.test_input_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000",
    "20200101000000000",
    "20200101000000001",
    "20200101000000002"
  ],
  "20200101000000001": [
    "20200101000000000"
  ],
  "20200101000000002": [
    "20200101000000003"
  ],
  "20200101000000003": [
    "20200101000000000"
  ]
})""")

        for info in wsb_exporter.run(self.test_input, self.test_output):
            pass

        with os.scandir(self.test_output) as entries:
            files = sorted(entries, key=lambda x: x.path)
        metas = []
        export_infos = []
        for file in files:
            with zipfile.ZipFile(file) as zh:
                with zh.open('meta.json') as fh:
                    metas.append(json.load(fh))
                with zh.open('export.json') as fh:
                    export_infos.append(json.load(fh))

        self.assertEqual(len(files), 7)

        # files are exported in depth-first order
        self.assertEqual(metas[0], {
            'id': '20200101000000000',
            'type': 'folder',
            'title': 'item0',
            })
        self.assertEqual(export_infos[0]['path'], [
            {'id': 'root', 'title': ''},
            ])

        self.assertEqual(metas[1], {
            'id': '20200101000000000',
            'type': 'folder',
            'title': 'item0',
            })
        self.assertEqual(export_infos[1]['path'], [
            {'id': 'root', 'title': ''},
            ])
        self.assertEqual(export_infos[1]['id'], export_infos[0]['id'])

        self.assertEqual(metas[2], {
            'id': '20200101000000001',
            'type': 'folder',
            'title': 'item1',
            })
        self.assertEqual(export_infos[2]['path'], [
            {'id': 'root', 'title': ''},
            ])

        self.assertEqual(metas[3], {
            'id': '20200101000000000',
            'type': 'folder',
            'title': 'item0',
            })
        self.assertEqual(export_infos[3]['path'], [
            {'id': 'root', 'title': ''},
            {'id': '20200101000000001', 'title': 'item1'},
            ])
        self.assertEqual(export_infos[3]['id'], export_infos[0]['id'])

        self.assertEqual(metas[4], {
            'id': '20200101000000002',
            'type': 'folder',
            'title': 'item2',
            })
        self.assertEqual(export_infos[4]['path'], [
            {'id': 'root', 'title': ''},
            ])

        self.assertEqual(metas[5], {
            'id': '20200101000000003',
            'type': 'folder',
            'title': 'item3',
            })
        self.assertEqual(export_infos[5]['path'], [
            {'id': 'root', 'title': ''},
            {'id': '20200101000000002', 'title': 'item2'},
            ])

        self.assertEqual(metas[6], {
            'id': '20200101000000000',
            'type': 'folder',
            'title': 'item0',
            })
        self.assertEqual(export_infos[6]['path'], [
            {'id': 'root', 'title': ''},
            {'id': '20200101000000002', 'title': 'item2'},
            {'id': '20200101000000003', 'title': 'item3'},
            ])
        self.assertEqual(export_infos[6]['id'], export_infos[0]['id'])

    def test_toc05(self):
        """Export first occurrence if singleton"""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder",
    "title": "item0"
  },
  "20200101000000001": {
    "type": "folder",
    "title": "item1"
  },
  "20200101000000002": {
    "type": "folder",
    "title": "item2"
  },
  "20200101000000003": {
    "type": "folder",
    "title": "item3"
  }
})""")
        with open(self.test_input_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000",
    "20200101000000000",
    "20200101000000001",
    "20200101000000002"
  ],
  "20200101000000001": [
    "20200101000000000"
  ],
  "20200101000000002": [
    "20200101000000003"
  ],
  "20200101000000003": [
    "20200101000000000"
  ]
})""")

        for info in wsb_exporter.run(self.test_input, self.test_output, singleton=True):
            pass

        with os.scandir(self.test_output) as entries:
            files = sorted(entries, key=lambda x: x.path)
        metas = []
        export_infos = []
        for file in files:
            with zipfile.ZipFile(file) as zh:
                with zh.open('meta.json') as fh:
                    metas.append(json.load(fh))
                with zh.open('export.json') as fh:
                    export_infos.append(json.load(fh))

        self.assertEqual(len(files), 4)

        # files are exported in depth-first order
        self.assertEqual(metas[0], {
            'id': '20200101000000000',
            'type': 'folder',
            'title': 'item0',
            })
        self.assertEqual(export_infos[0]['path'], [
            {'id': 'root', 'title': ''},
            ])

        self.assertEqual(metas[1], {
            'id': '20200101000000001',
            'type': 'folder',
            'title': 'item1',
            })
        self.assertEqual(export_infos[1]['path'], [
            {'id': 'root', 'title': ''},
            ])

        self.assertEqual(metas[2], {
            'id': '20200101000000002',
            'type': 'folder',
            'title': 'item2',
            })
        self.assertEqual(export_infos[2]['path'], [
            {'id': 'root', 'title': ''},
            ])

        self.assertEqual(metas[3], {
            'id': '20200101000000003',
            'type': 'folder',
            'title': 'item3',
            })
        self.assertEqual(export_infos[3]['path'], [
            {'id': 'root', 'title': ''},
            {'id': '20200101000000002', 'title': 'item2'},
            ])

    def test_toc06(self):
        """Export circular item but no children"""
        with open(self.test_input_meta, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.meta({
  "20200101000000000": {
    "type": "folder",
    "title": "item0"
  },
  "20200101000000001": {
    "type": "folder",
    "title": "item1"
  }
})""")
        with open(self.test_input_toc, 'w', encoding='UTF-8') as fh:
            fh.write("""\
scrapbook.toc({
  "root": [
    "20200101000000000"
  ],
  "20200101000000000": [
    "20200101000000001"
  ],
  "20200101000000001": [
    "20200101000000000"
  ]
})""")

        for info in wsb_exporter.run(self.test_input, self.test_output):
            pass

        with os.scandir(self.test_output) as entries:
            files = sorted(entries, key=lambda x: x.path)
        metas = []
        export_infos = []
        for file in files:
            with zipfile.ZipFile(file) as zh:
                with zh.open('meta.json') as fh:
                    metas.append(json.load(fh))
                with zh.open('export.json') as fh:
                    export_infos.append(json.load(fh))

        self.assertEqual(len(files), 3)

        # files are exported in depth-first order
        self.assertEqual(metas[0], {
            'id': '20200101000000000',
            'type': 'folder',
            'title': 'item0',
            })
        self.assertEqual(export_infos[0]['path'], [
            {'id': 'root', 'title': ''},
            ])

        self.assertEqual(metas[1], {
            'id': '20200101000000001',
            'type': 'folder',
            'title': 'item1',
            })
        self.assertEqual(export_infos[1]['path'], [
            {'id': 'root', 'title': ''},
            {'id': '20200101000000000', 'title': 'item0'},
            ])

        self.assertEqual(metas[2], {
            'id': '20200101000000000',
            'type': 'folder',
            'title': 'item0',
            })
        self.assertEqual(export_infos[2]['path'], [
            {'id': 'root', 'title': ''},
            {'id': '20200101000000000', 'title': 'item0'},
            {'id': '20200101000000001', 'title': 'item1'},
            ])
        self.assertEqual(export_infos[2]['id'], export_infos[0]['id'])

if __name__ == '__main__':
    unittest.main()



"""Generator of fulltext cache and/or static site pages.
"""
import os
import traceback
import shutil
import io
import zipfile
import mimetypes
import time
import re
import html
import copy
import itertools
import functools
from collections import namedtuple, UserDict
from urllib.parse import urlsplit, urljoin, quote, unquote
from datetime import datetime, timezone

import jinja2
from lxml import etree

from .host import Host
from .. import util
from ..util import Info
from .._compat import zip_stream
from .._compat.contextlib import nullcontext


class MutatingDict(UserDict):
    """Support adding during dict iteration.
    """
    def __init__(self, *args, **kwargs):
        self._keys = []

        # this calls __setitem__ internally
        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        if key not in self:
            self._keys.append(key)
        super().__setitem__(key, value)

    def __iter__(self):
        return iter(self._keys)

    def __delitem__(self, key):
        return NotImplemented


StaticIndexItem = namedtuple('StaticIndexItem',
    ['event', 'level', 'id', 'type', 'marked', 'title', 'url', 'icon', 'source', 'comment'])
StaticIndexItem.__new__.__defaults__ = (None, None, None, None, None, None, None, None)

class StaticSiteGenerator():
    """Main class for static site pages generation.
    """
    RESOURCES = {
        'icon/toggle.png': 'toggle.png',
        'icon/search.png': 'search.png',
        'icon/collapse.png': 'collapse.png',
        'icon/expand.png': 'expand.png',
        'icon/external.png': 'external.png',
        'icon/comment.png': 'comment.png',
        'icon/item.png': 'item.png',
        'icon/fclose.png': 'fclose.png',
        'icon/fopen.png': 'fopen.png',
        'icon/file.png': 'file.png',
        'icon/note.png': 'note.png',
        'icon/postit.png': 'postit.png',
        }
    ITEM_TYPE_ICON = {
        '': 'icon/item.png',
        'folder': 'icon/fclose.png',
        'file': 'icon/file.png',
        'image': 'icon/file.png',
        'note': 'icon/note.png',  # ScrapBook X notex
        'postit': 'icon/postit.png',  # ScrapBook X note
        }

    def __init__(self, book, *, locale=None,
            static_index=False, rss=False,
            ):
        self.host = book.host
        self.book = book
        self.static_index = static_index
        self.i18n = self.host.get_i18n(locale)

        self.rss = rss

        self.template_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(self.host.templates),
            autoescape=jinja2.select_autoescape(['html']),
            )
        self.template_env.globals.update({
            'format_string': util.format_string,
            })

        book.load_meta_files()
        book.load_toc_files()

    def run(self):
        yield Info('info', 'Generating static site pages...')

        # copy resource files
        for dst, src in self.RESOURCES.items():
            yield from self._generate_resource_file(src, dst)

        # generate static site pages
        index_kwargs = dict(
            title=self.book.name, i18n = self.i18n,
            rss=self.rss,
            data_dir = util.get_relative_url(self.book.data_dir, self.book.tree_dir),
            default_icons = self.ITEM_TYPE_ICON,
            meta_cnt=max(sum(1 for _ in self.book.iter_meta_files()), 1),
            toc_cnt=max(sum(1 for _ in self.book.iter_toc_files()), 1),
            )

        if self.static_index:
            yield from self._generate_page('index.html', 'static_index.html', filename='index',
                static_index=self._generate_static_index(), **index_kwargs,
                )

        yield from self._generate_page('map.html', 'static_map.html', filename='map',
            static_index=None, **index_kwargs,
            )

        yield from self._generate_page('frame.html', 'static_frame.html',
            title=self.book.name, i18n = self.i18n,
            )

        yield from self._generate_page('search.html', 'static_search.html',
            title=self.book.name, i18n = self.i18n,
            path=util.get_relative_url(self.book.top_dir, self.book.tree_dir),
            data_dir=util.get_relative_url(self.book.data_dir, self.book.top_dir),
            tree_dir=util.get_relative_url(self.book.tree_dir, self.book.top_dir),
            index=self.host.config['book'][self.book.id]['index'],
            )

    def _generate_resource_file(self, src, dst):
        yield Info('debug', f'Checking resource file "{dst}"')
        fsrc = self.host.get_static_file(src)
        fdst = os.path.normpath(os.path.join(self.book.tree_dir, dst))

        # check whether writing is required
        if os.path.isfile(fdst):
            if os.stat(fsrc).st_size == os.stat(fdst).st_size:
                if util.checksum(fsrc) == util.checksum(fdst):
                    yield Info('debug', f'Skipped resource file "{dst}" (up-to-date)')
                    return

        # save file
        yield Info('info', f'Generating resource file "{dst}"')
        try:
            os.makedirs(os.path.dirname(fdst), exist_ok=True)
            fsrc = self.host.get_static_file(src)
            self.book.backup(fdst)
            shutil.copyfile(fsrc, fdst)
        except OSError as exc:
            yield Info('error', f'Failed to create resource file "{dst}": [Errno {exc.args[0]}] {exc.args[1]}', exc=exc)

    def _generate_page(self, dst, tpl, **kwargs):
        yield Info('debug', f'Checking page "{dst}"')
        fsrc = io.BytesIO()
        fdst = os.path.normpath(os.path.join(self.book.tree_dir, dst))

        template = self.template_env.get_template(tpl)
        content = template.render(**kwargs)
        fsrc.write(content.encode('UTF-8'))

        # check whether writing is required
        if os.path.isfile(fdst):
            if fsrc.getbuffer().nbytes == os.stat(fdst).st_size:
                fsrc.seek(0)
                if util.checksum(fsrc) == util.checksum(fdst):
                    yield Info('debug', f'Skipped page "{dst}" (up-to-date)')
                    return

        # save file
        yield Info('info', f'Generating page "{dst}"')
        try:
            fsrc.seek(0)
            os.makedirs(os.path.dirname(fdst), exist_ok=True)
            self.book.backup(fdst)
            with open(fdst, 'wb') as fh:
                shutil.copyfileobj(fsrc, fh)
        except OSError as exc:
            yield Info('error', f'Failed to create page file "{dst}": [Errno {exc.args[0]}] {exc.args[1]}', exc=exc)

    def _generate_static_index(self):
        def get_class_text(classes, prefix=' '):
            if not classes:
                return ''

            c = html.escape(' '.join(classes))
            return f'{prefix}class="{c}"'

        def add_child_items(parent_id):
            nonlocal level

            try:
                toc = book.toc[parent_id]
            except KeyError:
                return

            toc = [id for id in toc if id in book.meta]
            if not toc:
                return

            yield StaticIndexItem('start-container', level)
            level += 1

            for id in toc:
                meta = book.meta[id]
                meta_type = meta.get('type', '')
                meta_index = meta.get('index', '')
                meta_title = meta.get('title', '')
                meta_source = meta.get('source', '')
                meta_icon = meta.get('icon', '')
                meta_comment = meta.get('comment', '')
                meta_marked = meta.get('marked', '')

                if meta_type != 'separator':
                    title = meta_title or id

                    if meta_type != 'folder':
                        if meta_type == 'bookmark' and meta_source:
                            href = meta_source
                        elif meta_index:
                            href = util.get_relative_url(os.path.join(book.data_dir, meta_index), book.tree_dir, path_is_dir=False)
                        else:
                            href = ''
                    else:
                        href = ''

                    # meta_icon is a URL
                    if meta_icon and not urlsplit(meta_icon).scheme:
                        # relative URL: tree_dir..index..icon
                        ref = util.get_relative_url(os.path.join(book.data_dir, os.path.dirname(meta_index)), book.tree_dir)
                        icon = ref + meta_icon
                    else:
                        icon = meta_icon

                else:
                    title = meta_title
                    href = ''
                    icon = ''

                yield StaticIndexItem('start', level, id, meta_type, meta_marked, title, href, icon, meta_source, meta_comment)

                # do not output children of a circular item
                if id not in id_chain:
                    level += 1
                    id_chain.add(id)
                    yield from add_child_items(id)
                    id_chain.remove(id)
                    level -= 1

                yield StaticIndexItem('end', level, id, meta_type, meta_marked, title, href, icon, meta_source, meta_comment)

            level -= 1
            yield StaticIndexItem('end-container', level)

        book = self.book
        level = 0
        id_chain = {'root'}
        yield from add_child_items('root')


class RssFeedGenerator():
    """Main class for RSS feed generation.
    """
    NS = 'http://www.w3.org/2005/Atom'

    def __init__(self, book, *, rss_root=None, item_count=50):
        self.book = book
        self.rss_root = rss_root.rstrip('/') + '/'
        self.item_count = item_count

        book.load_meta_files()
        book.load_toc_files()

    def run(self):
        yield Info('info', 'Generating RSS feed...')

        book = self.book
        rss_root = self.rss_root

        # RSS root must be an absolute URL
        u = urlsplit(rss_root)
        if not (u.scheme and u.netloc):
            yield Info('error', f'Invalid RSS root URL "{rss_root}"')
            return

        id_prefix = re.sub(r'/+$', '', f'urn:webscrapbook:{u.netloc}{u.path}')
        data_url = urljoin(rss_root, util.get_relative_url(book.data_dir, book.root))
        tree_url = urljoin(rss_root, util.get_relative_url(book.tree_dir, book.root))

        # get latest updated item entries
        entries = []
        for id, meta in book.meta.items():
            # show only items with content,
            # either with index or a bookmark with source
            if meta.get('type') in {'folder', 'separator'}:
                continue

            if meta.get('type') != 'bookmark' and meta.get('index'):
                pass
            elif meta.get('type') == 'bookmark' and meta.get('source'):
                pass
            else:
                continue

            entries.append({
                'id': id,
                'modify': meta.get('modify', meta.get('create', '')),
                'item': meta,
                })
        entries = sorted(entries, key=lambda d: d['modify'])
        entries = list(reversed(entries))[:self.item_count]

        # generate tree
        root = etree.XML(f'<feed xmlns="{self.NS}"></feed>'.encode('UTF-8'))

        elem = etree.SubElement(root, 'id')
        elem.text = id_prefix

        elem = etree.SubElement(root, 'link')
        elem.attrib['rel'] = 'self'
        elem.attrib['href'] = urljoin(tree_url, 'feed.atom')

        elem = etree.SubElement(root, 'link')
        elem.attrib['href'] = urljoin(tree_url, 'map.html')

        elem = etree.SubElement(root, 'title')
        elem.attrib['type'] = 'text'
        elem.text = book.name

        elem = etree.SubElement(root, 'updated')
        if entries:
            dt = util.id_to_datetime(entries[0]['modify'])
        else:
            dt = datetime.now(timezone.utc)
        elem.text = dt.strftime('%Y-%m-%dT%H:%M:%SZ')

        for entry in entries:
            entry_elem = etree.SubElement(root, 'entry')

            elem = etree.SubElement(entry_elem, 'id')
            elem.text = f'{id_prefix}:{quote(entry["id"])}'

            elem = etree.SubElement(entry_elem, 'link')
            if entry['item'].get('type') == 'bookmark':
                elem.attrib['href'] = entry['item']['source']
            else:
                elem.attrib['href'] = urljoin(data_url, quote(entry['item']['index']))

            elem = etree.SubElement(entry_elem, 'title')
            elem.attrib['type'] = 'text'
            elem.text = entry['item'].get('title', '')

            elem = etree.SubElement(entry_elem, 'published')
            dt = util.id_to_datetime(entry['item'].get('create', '')) or datetime.fromtimestamp(0, timezone.utc)
            elem.text = dt.strftime('%Y-%m-%dT%H:%M:%SZ')

            elem = etree.SubElement(entry_elem, 'updated')
            dt = util.id_to_datetime(entry['modify']) or datetime.fromtimestamp(0, timezone.utc)
            elem.text = dt.strftime('%Y-%m-%dT%H:%M:%SZ')

            elem = etree.SubElement(entry_elem, 'author')
            elem = etree.SubElement(elem, 'name')
            elem.text = 'Anonymous'

        # check whether writing is required
        fsrc = io.BytesIO(etree.tostring(root, xml_declaration=True, encoding='UTF-8'))
        fdst = os.path.normpath(os.path.join(self.book.tree_dir, 'feed.atom'))

        if os.path.isfile(fdst):
            if fsrc.getbuffer().nbytes == os.stat(fdst).st_size:
                fsrc.seek(0)
                if util.checksum(fsrc) == util.checksum(fdst):
                    yield Info('debug', 'Skipped RSS feed (up-to-date)')
                    return

        # save file
        yield Info('info', 'Generating RSS feed file "feed.atom"')
        try:
            fsrc.seek(0)
            os.makedirs(os.path.dirname(fdst), exist_ok=True)
            self.book.backup(fdst)
            with open(fdst, 'wb') as fh:
                shutil.copyfileobj(fsrc, fh)
        except OSError as exc:
            yield Info('error', f'Failed to create RSS feed file "feed.atom": [Errno {exc.args[0]}] {exc.args[1]}', exc=exc)


FulltextCacheItem = namedtuple('FulltextCacheItem', ['id', 'meta', 'index', 'indexfile', 'files_to_update'])

class FulltextCacheGenerator():
    """Main class for fulltext cache generation.
    """
    FULLTEXT_SPACE_REPLACER = functools.partial(re.compile(r'\s+').sub, ' ')
    FULLTEXT_EXCLUDE_TAGS = {
        'title', 'style', 'script',
        'frame', 'iframe',
        'object', 'applet',
        'audio', 'video',
        'canvas',
        'noframes', 'noscript', 'noembed',
        # 'parsererror',
        'svg', 'math',
        }
    URL_SAMPLE_LENGTH = 256

    def __init__(self, book, *, inclusive_frames=True, recreate=False):
        self.book = book
        self.inclusive_frames = inclusive_frames
        self.recreate = recreate
        self.cache_last_modified = 0

    def run(self, item_ids=None):
        """Update fulltext cache for item_ids

        Args:
            item_ids: a list of item IDs to update, invalid ones will be
                skipped. None to update all IDs.
        """
        yield Info('info', 'Generating fulltext cache...')

        book = self.book

        try:
            self.cache_last_modified = max(os.stat(f).st_mtime for f in book.iter_fulltext_files())
        except ValueError:
            # no fulltext file
            self.cache_last_modified = 0

        book.load_meta_files()
        book.load_toc_files()
        if self.recreate:
            book.fulltext = {}
            book_fulltext_orig = None
        else:
            book.load_fulltext_files()
            book_fulltext_orig = copy.deepcopy(book.fulltext)

        # generate cache for each item
        if item_ids:
            id_pool = dict.fromkeys(id for id in item_ids if id in book.meta or id in book.fulltext)
        else:
            id_pool = dict.fromkeys(itertools.chain(book.meta, book.fulltext))

        for id in id_pool:
            yield from self._cache_item(id)

        # update fulltext files
        if book.fulltext != book_fulltext_orig:
            # changed => save new files
            yield Info('info', f'Saving fulltext files...')
            book.save_fulltext_files()
        else:
            # no change => touch files to prevent falsely detected as outdated
            yield Info('info', f'Touching fulltext files...')
            for file in book.iter_fulltext_files():
                os.utime(file)

    def _cache_item(self, id):
        yield Info('debug', f'Checking item "{id}"')
        book = self.book

        # remove id if no meta
        meta = book.meta.get(id)
        if meta is None:
            yield Info('debug', f'Purging item "{id}" (missing metadata)')
            yield from self._delete_item(id)
            return

        # remove id if no index
        index = meta.get('index')
        if not index:
            yield Info('debug', f'Purging item "{id}" (no index)')
            yield from self._delete_item(id)
            return

        # remove id if no index file
        indexfile = os.path.join(book.data_dir, index)
        if not os.path.exists(indexfile):
            yield Info('debug', f'Purging item "{id}" (missing index file)')
            yield from self._delete_item(id)
            return

        # a mapping file path => status
        # status: True for a file to be checked; False for a file (mostly
        # an inclusive iframe) that is not available as inline,
        # (already added to cache or to be removed from cache)
        files_to_update = MutatingDict()

        item = FulltextCacheItem(id, meta, index, indexfile, files_to_update)
        yield from self._collect_files_to_update(item)
        yield from self._handle_files_to_update(item)

    def _delete_item(self, id):
        if id in self.book.fulltext:
            yield Info('info', f'Removing stale cache for "{id}".')
            del self.book.fulltext[id]

    def _collect_files_to_update(self, item):
        book = self.book
        id, meta, index, indexfile, files_to_update = item

        # create cache for this id if not exist yet
        if book.fulltext.get(id) is None:
            book.fulltext[id] = {}
        else:
            # unless newly created, presume no change if archive file not newer
            # than cache file, for better performance
            if util.is_archive(indexfile):
                if os.stat(indexfile).st_mtime <= self.cache_last_modified:
                    yield Info('debug', f'Skipped "{id}" (archive file older than cache)')
                    return

        # add index file(s) to update list
        try:
            for path in book.get_index_paths(index):
                yield Info('debug', f'Adding "{path}" of "{id}" to check list (from index)')
                files_to_update[path] = True
        except zipfile.BadZipFile:
            # MAFF file corrupted.
            # Skip adding index files.
            # Treat as no file exists and remove all indexes later on.
            yield Info('error', f'Archive file for "{id}" is corrupted')

        # add files in cache to update list
        for path in book.fulltext[id]:
            yield Info('debug', f'Adding "{path}" of "{id}" to check list (from cache)')
            files_to_update[path] = True

    def _handle_files_to_update(self, item):
        def report_update():
            nonlocal has_update
            if has_update:
                return
            if book.fulltext[id]:
                yield Info('info', f'Updating cache for "{id}"...')
            else:
                yield Info('info', f'Generating cache for "{id}"...')
            has_update = True

        book = self.book
        id, meta, index, indexfile, files_to_update = item
        has_update = False

        for path in files_to_update:
            yield Info('debug', f'Checking "{path}" of "{id}"')
            # remove from cache if marked False
            if not files_to_update[path]:
                yield Info('debug', f'Purging "{path}" of "{id}" (inlined)')
                if path in book.fulltext[id]:
                    yield from report_update()
                    del book.fulltext[id][path]
                continue

            # mark False to prevent added otherwhere
            files_to_update[path] = False

            mtime = yield from self._get_mtime(item, path)
            if mtime is None:
                # path not exist => delete from cache
                yield Info('debug', f'Purging "{path}" of "{id}" (file not exist)')
                if path in book.fulltext[id]:
                    yield from report_update()
                    del book.fulltext[id][path]
                continue

            # skip update if the file is not newer
            # - A file hasn't been cached may be newly refrenced by another
            #   updated file, and thus needs update even if it's mtime is not
            #   newer.
            if path in book.fulltext[id] and mtime <= self.cache_last_modified:
                yield Info('debug', f'Skipped "{path}" of "{id}" (file older than cache)')
                continue

            yield from report_update()

            # set updated fulltext
            yield Info('debug', f'Generating cache for "{path}" of "{id}"')
            try:
                fulltext = yield from self._get_fulltext_cache(item, path)
            except Exception as exc:
                fulltext = ''
                traceback.print_exc()
                yield Info('error', f'Failed to generate cache for "{id}" ({path}): {exc}', exc=exc)

            if fulltext is not None:
                book.fulltext[id][path] = {
                    'content': fulltext,
                    }
            else:
                try:
                    del book.fulltext[id][path]
                except KeyError:
                    pass

    def _get_mtime(self, item, path):
        if util.is_archive(item.index):
            try:
                zh = zipfile.ZipFile(os.path.join(self.book.data_dir, item.index))
            except zipfile.BadZipFile as exc:
                yield Info('error', f'Failed to open zip file "{item.index}" for "{item.id}": {exc}', exc=exc)
                return None
            except (FileNotFoundError, IsADirectoryError, NotADirectoryError) as exc:
                yield Info('error', f'Failed to open zip file "{item.index}" for "{item.id}": [Errno {exc.args[0]}] {exc.args[1]}', exc=exc)
                return None

            try:
                with zh as zh:
                    info = zh.getinfo(path)
                    return util.zip_timestamp(info)
            except KeyError:
                return None
            except Exception as exc:
                yield Info('error', f'Failed to access in-zip-file for "{path}" of "{item.id}": {exc}', exc=exc)
                return None

        file = os.path.join(self.book.data_dir, os.path.dirname(item.index), path)
        try:
            return os.stat(file).st_mtime
        except (FileNotFoundError, IsADirectoryError, NotADirectoryError):
            return None
        except OSError as exc:
            yield Info('error', f'Failed to access file for "{path}" of "{item.id}": [Errno {exc.args[0]}] {exc.args[1]}', exc=exc)
            return None

    def _open_file(self, item, path):
        if util.is_archive(item.index):
            try:
                zh = zipfile.ZipFile(os.path.join(self.book.data_dir, item.index))
            except zipfile.BadZipFile as exc:
                yield Info('error', f'Failed to open zip file "{item.index}" for "{item.id}": {exc}', exc=exc)
                return None
            except (FileNotFoundError, IsADirectoryError, NotADirectoryError) as exc:
                yield Info('error', f'Failed to open zip file "{item.index}" for "{item.id}": [Errno {exc.args[0]}] {exc.args[1]}', exc=exc)
                return None

            try:
                with zh as zh:
                    return zh.open(path)
            except KeyError:
                return None
            except Exception as exc:
                yield Info('error', f'Failed to open in-zip-file for "{path}" of "{item.id}": {exc}', exc=exc)
                return None

        file = os.path.join(self.book.data_dir, os.path.dirname(item.index), path)
        try:
            return open(file, 'rb')
        except (FileNotFoundError, IsADirectoryError, NotADirectoryError):
            return None
        except OSError as exc:
            yield Info('error', f'Failed to open file for "{path}" of "{item.id}": [Errno {exc.args[0]}] {exc.args[1]}', exc=exc)
            return None

    def _get_fulltext_cache(self, item, path):
        fh = yield from self._open_file(item, path)
        if not fh:
            yield Info('debug', f'Skipped "{path}" of "{item.id}" (file not exist or accessible)')
            return None

        fh = zip_stream(fh)
        try:
            mime, _ = mimetypes.guess_type(path)
            return (yield from self._get_fulltext_cache_for_fh(item, path, fh, mime))
        finally:
            fh.close()

    def _get_fulltext_cache_for_fh(self, item, path, fh, mime):
        if not mime:
            yield Info('debug', f'Skipped "{path}" of "{item.id}" (unknown type)')
            return None

        if util.mime_is_html(mime):
            return (yield from self._get_fulltext_cache_html(item, path, fh))

        if mime.startswith('text/'):
            return (yield from self._get_fulltext_cache_txt(item, path, fh))

        yield Info('debug', f'Skipped "{path}" of "{item.id}" ("{mime}" not supported)')
        return None

    def _get_fulltext_cache_html(self, item, path, fh):
        def get_relative_file_path(url):
            # skip when inside a data URL page (can't resolve)
            if path is None:
                return None

            urlparts = urlsplit(url)

            # skip absolute URLs
            if urlparts.scheme != '':
                return None

            if urlparts.netloc != '':
                return None

            if urlparts.path.startswith('/'):
                return None

            base = get_relative_file_path.base = getattr(get_relative_file_path, 'base', 'file:///!/')
            ref = get_relative_file_path.ref = getattr(get_relative_file_path, 'ref', urljoin(base, quote(path)))
            target = urljoin(ref, urlparts.path)

            # skip if URL contains '..'
            if not target.startswith(base):
                return None

            target = unquote(target)

            # ignore referring self
            if target == ref:
                return None

            target = target[len(base):]

            return target

        def add_datauri_content(url):
            try:
                data = util.parse_datauri(url)
            except util.DataUriMalformedError as exc:
                yield Info('error', f'Skipped malformed data URL "{url[:self.URL_SAMPLE_LENGTH]}": {exc}', exc=exc)
                return
            fh = io.BytesIO(data.bytes)
            fulltext = yield from self._get_fulltext_cache_for_fh(item, None, fh, data.mime)
            if fulltext:
                results.append(fulltext)

        yield Info('debug', f'Retrieving HTML content for "{path}" of "{item.id}"')

        # Seek for the correct charset (encoding).
        # If a charset is not specified, lxml may select a wrong encoding for
        # the entire document if there is text before first meta charset.
        # Priority: BOM > meta charset > item charset > assume UTF-8
        charset = util.sniff_bom(fh)
        if charset:
            # lxml does not accept "UTF-16-LE" or so, but can auto-detect
            # encoding from BOM if encoding is None
            # ref: https://bugs.launchpad.net/lxml/+bug/1463610
            charset = None
            fh.seek(0)
        else:
            charset = util.get_charset(fh) or item.meta.get('charset') or 'UTF-8'
            charset = util.fix_codec(charset)
            fh.seek(0)

        results = []
        has_instant_redirect = False
        for time_, url, context in util.iter_meta_refresh(fh):
            if time_ == 0 and not context:
                has_instant_redirect = True

            if not url:
                continue

            if context and any(c in util.META_REFRESH_FORBID_TAGS for c in context):
                continue

            if url.startswith('data:'):
                yield from add_datauri_content(url)
            else:
                target = get_relative_file_path(url)
                if target and target not in item.files_to_update:
                    yield Info('debug', f'Adding "{target}" of "{item.id}" to check list (from <meta>)')
                    item.files_to_update[target] = True

        # Add data URL content of meta refresh targets to fulltext index if the
        # page has an instant meta refresh.
        if has_instant_redirect:
            return self.FULLTEXT_SPACE_REPLACER(' '.join(results)).strip()

        # add main content
        # Note: adding elem.text at start event or elem.tail at end event is
        # not reliable as the parser hasn't load full content of text or tail
        # at that time yet.
        # @TODO: better handle content
        # (no space between inline nodes, line break between block nodes, etc.)
        fh.seek(0)
        exclusion_stack = []
        for event, elem in etree.iterparse(fh, html=True, events=('start', 'end'),
                remove_comments=True, encoding=charset):
            if event == 'start':
                # skip if we are in an excluded element
                if exclusion_stack:
                    continue

                # Add last text before starting of this element.
                prev = elem.getprevious()
                attr = 'tail'
                if prev is None:
                    prev = elem.getparent()
                    attr = 'text'

                if prev is not None:
                    text = getattr(prev, attr)
                    if text:
                        results.append(text)
                        setattr(prev, attr, None)

                if elem.tag in ('a', 'area'):
                    # include linked pages in fulltext index
                    try:
                        url = elem.attrib['href']
                    except KeyError:
                        pass
                    else:
                        if url.startswith('data:'):
                            yield from add_datauri_content(url)
                        else:
                            target = get_relative_file_path(url)
                            if target and target not in item.files_to_update:
                                yield Info('debug', f'Adding "{target}" of "{item.id}" to check list (from <{elem.tag}>)')
                                item.files_to_update[target] = True

                elif elem.tag in ('iframe', 'frame'):
                    # include frame page in fulltext index
                    try:
                        url = elem.attrib['src']
                    except KeyError:
                        pass
                    else:
                        if url.startswith('data:'):
                            yield from add_datauri_content(url)
                        else:
                            target = get_relative_file_path(url)
                            if target:
                                if self.inclusive_frames:
                                    # Add frame content to the current page
                                    # content if the targeted file hasn't
                                    # been indexed.
                                    if item.files_to_update.get(target) is not False:
                                        yield Info('debug', f'Caching "{target}" of "{item.id}" as inline (from <{elem.tag}>)')
                                        item.files_to_update[target] = False
                                        fulltext = yield from self._get_fulltext_cache(item, target)
                                        if fulltext:
                                            results.append(fulltext)
                                else:
                                    if target not in item.files_to_update:
                                        yield Info('debug', f'Adding "{target}" of "{item.id}" to check list (from <{elem.tag}>)')
                                        item.files_to_update[target] = True

                # exclude everything inside certain tags
                if elem.tag in self.FULLTEXT_EXCLUDE_TAGS:
                    exclusion_stack.append(elem)
                    continue

            elif event == 'end':
                # Add last text before ending of this element.
                if not exclusion_stack:
                    try:
                        prev = elem[-1]
                        attr = 'tail'
                    except IndexError:
                        prev = elem
                        attr = 'text'

                    if prev is not None:
                        text = getattr(prev, attr)
                        if text:
                            results.append(text)
                            setattr(prev, attr, None)

                # stop exclusion at the end of an excluding element
                try:
                    if elem is exclusion_stack[-1]:
                        exclusion_stack.pop()
                except IndexError:
                    pass

                # clean up to save memory
                # remember to keep tail
                try:
                    elem.clear(keep_tail=True)
                except TypeError:
                    # keep_tail is supported since lxml 4.4.0
                    pass
                while elem.getprevious() is not None:
                    try:
                        del elem.getparent()[0]
                    except TypeError:
                        # broken html may generate extra root elem
                        break

        return self.FULLTEXT_SPACE_REPLACER(' '.join(results)).strip()

    def _get_fulltext_cache_txt(self, item, path, fh):
        yield Info('debug', f'Retrieving text content for "{path}" of "{item.id}"')
        charset = util.sniff_bom(fh) or item.meta.get('charset') or 'UTF-8'
        charset = util.fix_codec(charset)
        text = fh.read().decode(charset, errors='replace')
        return self.FULLTEXT_SPACE_REPLACER(text).strip()


def generate(root, book_ids=None, item_ids=None, *,
        config=None, no_lock=False, no_backup=False,
        fulltext=True, inclusive_frames=True, recreate=False,
        static_site=False, static_index=False, locale=None,
        rss_root=None, rss_item_count=50):
    start = time.time()

    host = Host(root, config)

    # cache all book_ids if none specified
    if not book_ids:
        book_ids = list(host.books)

    ts = None
    avail_book_ids = set(host.books)
    for book_id in book_ids:
        # skip invalid book ID
        if book_id not in avail_book_ids:
            yield Info('warn', f'Skipped invalid book "{book_id}".')
            continue

        yield Info('debug', f'Loading book "{book_id}".')

        try:
            book = host.books[book_id]

            if book.no_tree:
                yield Info('info', f'Skipped book "{book_id}" ("{book.name}") (no_tree).')
                continue

            yield Info('info', f'Caching book "{book_id}" ({book.name}).')
            lh = nullcontext() if no_lock else book.get_tree_lock().acquire()
            with lh:
                if not no_backup:
                    # use same timestamp for all books
                    ts = ts or util.datetime_to_id()
                    book.init_backup(ts)
                    yield Info('info', f'Prepared backup at "{book.get_subpath(book.backup_dir)}".')

                try:
                    if fulltext:
                        generator = FulltextCacheGenerator(
                            book,
                            inclusive_frames=inclusive_frames,
                            recreate=recreate,
                            )
                        yield from generator.run(item_ids)

                    if static_site:
                        generator = StaticSiteGenerator(
                            book,
                            static_index=static_index,
                            locale=locale, rss=bool(rss_root),
                            )
                        yield from generator.run()

                    if rss_root:
                        generator = RssFeedGenerator(
                            book,
                            rss_root=rss_root,
                            item_count=rss_item_count,
                            )
                        yield from generator.run()
                finally:
                    if not no_backup:
                        book.init_backup(False)

        except Exception as exc:
            traceback.print_exc()
            yield Info('critical', str(exc), exc=exc)
        else:
            yield Info('info', 'Done.')

        yield Info('info', '----------------------------------------------------------------------')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



import os
import shutil
import zipfile
import traceback
import time
import html

from ... import util
from ...util import Info
from ..host import Host
from .. import book as wsb_book


class Converter:
    def __init__(self, input, output, book_id='', prefix=True):
        self.input = input
        self.output = output
        self.book_id = book_id
        self.book = None
        self.prefix = prefix

        self.used_paths = set()

    def run(self):
        host = Host(self.input)

        try:
            self.book = host.books[self.book_id]
        except KeyError as exc:
            raise RuntimeError(f'book "{self.book_id}" does not exist') from exc

        self.book.load_meta_files()
        self.book.load_toc_files()

        id_chain = set()
        path_chain = []
        ref_ids = self.book.toc.get('root', [])
        idx_len = len(str(len(ref_ids)))
        for idx, ref_id in enumerate(ref_ids):
            yield from self._export_item(ref_id, idx, idx_len, id_chain, path_chain)

    def _export_item(self, id, idx, idx_len, id_chain, path_chain):
        if id not in self.book.meta or id in wsb_book.Book.SPECIAL_ITEM_ID:
            return

        yield Info('debug', f'Exporting item "{id}"')

        meta = self.book.meta[id]
        type = meta.get('type', '')
        prefix = f'{idx + 1:0{idx_len}d}-' if self.prefix else ''
        basename = self.book.meta[id].get('title') or (id if type != 'separator' else '----')

        path = os.path.join(self.output, *path_chain, util.crop(prefix + util.validate_filename(basename), 128))
        path = self._get_unique_path(path)
        self.used_paths.add(path)

        if type == 'folder':
            dst = path
            subpath = os.path.relpath(dst, self.output)
            yield Info('info', f'Exporting "{id}" to "{subpath}"...')
            try:
                os.makedirs(dst, exist_ok=True)
            except OSError as exc:
                yield Info('error', f'Failed to export "{id}" to "{subpath}": {exc}')

        elif type == 'separator':
            if self.prefix:
                dst = path + '.-'
                subpath = os.path.relpath(dst, self.output)
                yield Info('info', f'Exporting "{id}" to "{subpath}"...')
                try:
                    os.makedirs(os.path.dirname(dst), exist_ok=True)
                    with open(dst, 'wb') as fh:
                        pass
                except OSError as exc:
                    yield Info('error', f'Failed to export "{id}" to "{subpath}": {exc}')

        elif type == 'bookmark':
            source = meta.get('source')
            if source:
                dst = path + '.htm'
                subpath = os.path.relpath(dst, self.output)
                yield Info('info', f'Exporting "{id}" to "{subpath}"...')
                try:
                    os.makedirs(os.path.dirname(dst), exist_ok=True)
                    with open(dst, 'w', encoding='UTF-8') as fh:
                        fh.write(f'<!DOCTYPE html><meta charset="UTF-8"><meta http-equiv="refresh" content="0;url={html.escape(source)}">')
                except OSError as exc:
                    yield Info('error', f'Failed to export "{id}" to "{subpath}": {exc}')

        else:
            index = meta.get('index', '')
            if index:
                if index.endswith('/index.html'):
                    src = os.path.normpath(os.path.dirname(os.path.join(self.book.data_dir, index)))
                    ext = '.htd'
                else:
                    src = os.path.normpath(os.path.join(self.book.data_dir, index))
                    _, ext = os.path.splitext(src)

                    if not ext:
                        # prevent conflict with folder
                        ext = '._'

                dst = path + ext
                subpath = os.path.relpath(dst, self.output)
                yield Info('info', f'Exporting "{id}" to "{subpath}"...')
                try:
                    os.makedirs(os.path.dirname(dst), exist_ok=True)
                    try:
                        shutil.copytree(src, dst)
                    except NotADirectoryError:
                        shutil.copy2(src, dst)
                except OSError as exc:
                    yield Info('error', f'Failed to export "{id}" to "{subpath}": {exc}')

        # do not add descendants for a recursive item
        if id in id_chain:
            return

        id_chain.add(id)
        path_chain.append(os.path.basename(path))
        ref_ids = self.book.toc.get(id, [])
        idx_len = len(str(len(ref_ids)))
        for idx, ref_id in enumerate(ref_ids):
            yield from self._export_item(ref_id, idx, idx_len, id_chain, path_chain)
        path_chain.pop()
        id_chain.remove(id)

    def _get_unique_path(self, path):
        i = 0
        prefix = path
        while path in self.used_paths:
            i += 1
            path = f'{prefix}({i})'
        return path


def run(input, output, book_id='', prefix=True):
    start = time.time()
    yield Info('info', 'conversion mode: WebScrapBook --> hierarchial files')
    yield Info('info', f'input directory: {os.path.abspath(input)}')
    yield Info('info', f'output directory: {os.path.abspath(output)}')
    yield Info('info', f'book ID: "{book_id}"')
    yield Info('info', f'prefix: {prefix}')
    yield Info('info', '')

    try:
        conv = Converter(input, output, book_id=book_id, prefix=prefix)
        yield from conv.run()
    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
    else:
        yield Info('info', 'Done.')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



import os
import shutil
import traceback
import re
import time
import functools
from datetime import datetime
from urllib.parse import unquote
from urllib.request import pathname2url

from lxml import etree

from ... import WSB_DIR, WSB_CONFIG
from ... import util
from ...util import Info
from ..host import Host
from .. import book as wsb_book


RDF = '{http://www.w3.org/1999/02/22-rdf-syntax-ns#}'
NS1 = '{http://amb.vis.ne.jp/mozilla/scrapbook-rdf#}'
NC = '{http://home.netscape.com/NC-rdf#}'
RES_PROTOCOL_BASE = 'resource://scrapbook/'
MOZ_ICON_BASE = 'moz-icon://'

REGEX_ID = re.compile(r'\d{14}')
REGEX_RDF_ID = re.compile(r'urn:scrapbook:item(\d{14})')
REGEX_RDF_SEQ_ID = re.compile(r'urn:scrapbook:(?:item(\d{14})|(root))')

LEGACY_TYPE_MAP = {
    "note": "postit",
    "notex": "note",
    }

PRUNE_FILES = {
    'backup',
    'tree',
    'scrapbook.rdf',
    'cache.rdf',
    'folders.txt',
    'collection.html',
    'combine.html',
    'combine.css',
    'note.html',
    'search.html',
    'sitemap.xsl',
    'note_template.html',
    'notex_template.html',
    }


class LegacyBook:
    """Main legacy scrapbook book controller.
    """
    def __init__(self, root):
        self.root = os.path.realpath(root)
        self.meta = {}
        self.toc = {'root': []}
        self.rdf_mtime = 0
        self.sitemaps = {}
        self.modifies = {}

    def load(self):
        yield Info('info', 'Loading "scrapbook.rdf"...')
        yield from self._load_rdf()
        yield Info('info', 'Inspecting data directory...')
        yield from self._load_data_dir()

    def _load_rdf(self):
        rdf_file = os.path.join(self.root, 'scrapbook.rdf')
        try:
            self.rdf_mtime = os.stat(rdf_file).st_mtime
            with open(rdf_file, 'rb') as fh:
                yield from self._parse_rdf(fh)
        except OSError as exc:
            raise RuntimeError(f'Unable to load "scrapbook.rdf": [Errno {exc.args[0]}] {exc.args[1]}') from exc
        except etree.XMLSyntaxError as exc:
            raise RuntimeError(f'Malformed "scrapbook.rdf": {exc.args[0]}') from exc

    def _parse_rdf(self, fh):
        for event, elem in etree.iterparse(fh, events=('end',),
                remove_comments=True, encoding='UTF-8',
                tag=(f'{RDF}Description', f'{NC}BookmarkSeparator', f'{RDF}Seq'),
                ):
            yield Info('debug', f'Inspecting element: {elem.tag} ({elem.attrib.get(f"{RDF}about")})')
            if elem.tag in {f'{RDF}Description', f'{NC}BookmarkSeparator'}:
                rid = elem.attrib[f'{RDF}about']
                rid_match = REGEX_RDF_ID.match(rid)
                if not rid_match:
                    continue

                id = elem.attrib.get(f'{NS1}id') or rid_match.group(1)

                meta = {'id': id}
                for attr, value in elem.attrib.items():
                    if attr.startswith(NS1) and attr != f'{NS1}id':
                        meta[attr[len(NS1):]] = value
                self.meta.setdefault(id, {}).update(meta)

            elif elem.tag == f'{RDF}Seq':
                rid = elem.attrib[f'{RDF}about']
                rid_match = REGEX_RDF_SEQ_ID.match(rid)
                if not rid_match:
                    continue

                id = rid_match.group(1) or rid_match.group(2)

                for child in elem:
                    if child.tag == f'{RDF}li':
                        ref_rid = child.attrib[f'{RDF}resource']
                        ref_rid_match = REGEX_RDF_ID.match(ref_rid)
                        if not ref_rid_match:
                            continue

                        ref_id = ref_rid_match.group(1)
                        self.toc.setdefault(id, []).append(ref_id)

            # clean up to save memory
            elem.clear()
            while elem.getprevious() is not None:
                del elem.getparent()[0]

    def _load_data_dir(self, load_index_dat=False):
        try:
            dirs = os.scandir(os.path.join(self.root, 'data'))
        except (FileNotFoundError, NotADirectoryError):
            yield Info('debug', 'Skipped inspecting data directory (not found)')
            return
        except OSError as exc:
            raise RuntimeError(f'Failed to scan data directory: [Errno {exc.args[0]}] {exc.args[1]}') from exc

        with dirs as dirs:
            for dir in dirs:
                # check name as ID
                id = dir.name
                yield Info('debug', f'Inspecting item folder "{id}"')
                match_id = REGEX_ID.match(id)
                if not match_id:
                    yield Info('warn', f'Skipped item folder "{id}" (invalid ID)')
                    continue

                # check index.html
                index_file = os.path.join(dir, 'index.html')
                if not os.path.isfile(index_file):
                    yield Info('debug', f'Skipped item folder "{id}" (missing "index.html")')
                    continue

                # record mtime of index.html
                ts = os.stat(index_file).st_mtime
                mtime = datetime.fromtimestamp(ts)
                modify = util.datetime_to_id(mtime)
                self.modifies[id] = modify

                # record seen sitemap.xml
                sitemap_file = os.path.join(dir, 'sitemap.xml')
                if os.path.isfile(sitemap_file):
                    self.sitemaps[id] = True

                # load index.dat
                if load_index_dat:
                    yield Info('debug', f'Checking "index.dat" for "{id}"')
                    index_dat_file = os.path.join(dir, 'index.dat')
                    try:
                        assert os.stat(index_dat_file).st_mtime > self.rdf_mtime
                    except (FileNotFoundError, IsADirectoryError, NotADirectoryError, AssertionError):
                        continue
                    except OSError as exc:
                        yield Info('error', f'Failed to read "index.dat" for "{id}": [Errno {exc.args[0]}] {exc.args[1]}', exc=exc)
                        continue

                    yield Info('info', f'Reading metadata from newer "index.dat" for "{id}"')
                    try:
                        with open(index_dat_file, encoding='UTF-8') as fh:
                            meta = {}
                            for line in fh:
                                line = line.rstrip('\n')
                                key, _, value = line.partition('\t')
                                if not value:
                                    continue
                                meta[key] = value
                            self.meta.setdefault(id, {}).update(meta)
                    except OSError as exc:
                        yield Info('error', f'Failed to read "index.dat" for "{id}": [Errno {exc.args[0]}] {exc.args[1]}', exc=exc)
                        continue


class Converter:
    def __init__(self, input, output, * , no_backup=False):
        self.input = input
        self.output = output
        self.no_backup = no_backup

    def run(self):
        fsrc = os.path.normpath(os.path.join(__file__, '..', '..', '..', 'resources', 'config.ini'))
        fdst = os.path.normpath(os.path.join(self.output, WSB_DIR, WSB_CONFIG))

        yield Info('info', 'Generating directory...')
        os.makedirs(os.path.dirname(fdst), exist_ok=True)

        if not os.path.isfile(fdst):
            yield Info('info', 'Generating config file...')
            try:
                shutil.copyfile(fsrc, fdst)
            except OSError as exc:
                yield Info('error', f'Unable to generate config file: {exc}', exc=exc)

        yield Info('info', 'Setting up new scrapbook...')
        host = Host(self.output)
        book = host.books['']
        book.load_meta_files()
        book.load_toc_files()

        if not self.no_backup:
            book.init_backup()

        try:
            yield Info('info', 'Loading legacy scrapbook data...')
            book0 = LegacyBook(self.input)
            yield from book0.load()

            yield Info('info', 'Calculating tree data...')
            yield from self._merge_meta(book, book0)
            yield from self._merge_toc(book, book0)

            yield Info('info', 'Copying data files...')
            yield from self._copy_data_files(book, book0)
            yield from self._convert_data_files(book, book0)

            yield Info('info', 'Saving tree files...')
            book.save_meta_files()
            book.save_toc_files()
        finally:
            if not self.no_backup:
                book.init_backup(False)

    def _merge_meta(self, book, book0):
        for id0, meta0 in book0.meta.items():
            yield Info('debug', f'Inspecting item metadata for "{id0}"')
            id = id0
            meta = meta0.copy()
            meta_new = book.meta.setdefault(id, book.DEFAULT_META.copy())

            # meta['id']
            # omit id field in meta
            try:
                del meta['id']
            except KeyError:
                pass

            # meta['type'], meta['marked']
            meta['type'] = LEGACY_TYPE_MAP.get(meta.get('type'), meta.get('type', ''))
            if meta['type'] == 'marked':
                # if sitemap exists, type should be site
                meta['type'] = 'site' if id in book0.sitemaps else ''
                meta['marked'] = True

            # meta['index']
            if meta['type'] not in wsb_book.Book.TYPES_OPTIONAL_INDEX:
                meta['index'] = f'{id}/index.html'

            # meta['create']
            # fallback to id
            if meta.get('create'):
                meta['create'] = util.datetime_to_id(util.id_to_datetime_legacy(meta['create']))
            else:
                meta['create'] = util.datetime_to_id(util.id_to_datetime_legacy(id))

            # meta['modify']
            # fallback to create (and then id)
            if meta.get('modify'):
                meta['modify'] = util.datetime_to_id(util.id_to_datetime_legacy(meta['modify']))
            else:
                meta['modify'] = meta['create']

            # take mtime of index.html if newer
            try:
                mtime = book0.modifies[id]
            except KeyError:
                pass
            else:
                if mtime > meta['modify']:
                    yield Info('debug', f'Taking newer timestamp from "index.html" as modify time for "{id}"')
                    meta['modify'] = mtime

            # meta['icon']
            # resolve resource://scrapbook/* and moz-icon://*
            if meta.get('icon', '').startswith(RES_PROTOCOL_BASE):
                src = os.path.normpath(os.path.join(book.data_dir, id))
                dst = os.path.normpath(os.path.join(book.top_dir, unquote(meta['icon'][len(RES_PROTOCOL_BASE):])))
                rel_path = os.path.relpath(dst, src)
                meta['icon'] = pathname2url(rel_path)  # this quotes URL
            elif meta.get('icon', '').startswith(MOZ_ICON_BASE):
                meta['icon-moz'] = meta['icon']
                meta['icon'] = ''

            # meta['comment']
            try:
                if meta['comment']:
                    meta['comment'] = meta['comment'].replace(' __BR__ ', '\n')
            except KeyError:
                pass

            # meta['charset']
            try:
                if meta['chars']:
                    meta['charset'] = meta['chars']
                del meta['chars']
            except KeyError:
                pass

            # meta['lock']
            try:
                if meta['lock']:
                    meta['locked'] = True
                del meta['lock']
            except KeyError:
                pass

            # meta['folder']
            # this should not appear in scrapbook.rdf normally
            try:
                del meta['folder']
            except KeyError:
                pass

            # meta['exported']
            # this should not appear in scrapbook.rdf normally
            try:
                del meta['exported']
            except KeyError:
                pass

            # meta['container']
            # this should not appear in scrapbook.rdf normally
            try:
                del meta['container']
            except KeyError:
                pass

            # merge and remove None values
            meta_new.update(meta)
            for k, v in list(meta_new.items()):
                if v is None:
                    del meta_new[k]

    def _merge_toc(self, book, book0):
        book.toc.update(book0.toc)
        return
        yield

    def _copy_data_files(self, book, book0):
        with os.scandir(self.input) as dirs:
            for src in dirs:
                if src.name == WSB_DIR:
                    yield Info('warn', f'Skipped copying special directory "{WSB_DIR}"')
                    continue

                if src.name in PRUNE_FILES:
                    if self.no_backup:
                        yield Info('debug', f'Skipped legacy scrapbook entry "{src.name}"')
                        continue
                    else:
                        yield Info('debug', f'Backup legacy scrapbook entry "{src.name}"')
                        book.backup(src, base=self.input)
                        continue
                else:
                    dst = os.path.join(self.output, src.name)

                try:
                    shutil.copytree(src, dst)
                except NotADirectoryError:
                    shutil.copy2(src, dst)

    def _convert_data_files(self, book, book0):
        for id, meta in book.meta.items():
            if meta['type'] == 'postit':
                yield Info('debug', f'Converting data file for "{id}": type={meta["type"]}')
                index_file = os.path.normpath(os.path.join(book.data_dir, meta['index']))
                try:
                    content = book.load_note_file(index_file)
                except OSError:
                    pass
                else:
                    book.save_note_file(index_file, content)


def run(input, output, * , no_backup=False):
    start = time.time()
    yield Info('info', 'conversion mode: ScrapBook --> WebScrapBook')
    yield Info('info', f'input directory: {os.path.abspath(input)}')
    yield Info('info', f'output directory: {os.path.abspath(output)}')
    yield Info('info', f'no-backup: {no_backup}')
    yield Info('info', '')

    try:
        conv = Converter(input, output, no_backup=no_backup)
        yield from conv.run()
    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
    else:
        yield Info('info', 'Done.')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



"""A backend toolkit for management of WebScrapBook collection.
"""
import sys
import os
from configparser import ConfigParser
from collections import OrderedDict
from copy import deepcopy
import mimetypes
import re

__all__ = ['WSB_EXTENSION_MIN_VERSION', 'WSB_USER_DIR', 'WSB_USER_CONFIG', 'WSB_DIR', 'WSB_CONFIG', 'config']

__package_name__ = 'webscrapbook'
__version__ = '0.32.1'
__author__ = 'Danny Lin'
__author_email__ = 'danny0838@gmail.com'
__homepage__ = 'https://github.com/danny0838/PyWebScrapBook'
__license__ = 'MIT'

WSB_EXTENSION_MIN_VERSION = '0.79.0'
WSB_USER_DIR = os.path.join(os.path.expanduser('~'), '.config', 'wsb')  # affected by $HOME
WSB_USER_CONFIG = os.path.join(os.path.expanduser('~'), '.wsbconfig')  # affected by $HOME
WSB_DIR = os.environ.get('WSB_DIR') or '.wsb'
WSB_CONFIG = 'config.ini'

mimetypes.add_type("application/rss+xml", ".rss")
mimetypes.add_type("application/atom+xml", ".atom")
mimetypes.add_type("text/markdown", ".md")
mimetypes.add_type("text/markdown", ".mkd")
mimetypes.add_type("text/markdown", ".mkdn")
mimetypes.add_type("text/markdown", ".mdwn")
mimetypes.add_type("text/markdown", ".mdown")
mimetypes.add_type("text/markdown", ".markdown")
mimetypes.add_type("application/html+zip", ".htz")
mimetypes.add_type("application/x-maff", ".maff")
mimetypes.add_type("application/wsba+zip", ".wsba")

# seems to be set to x-zip-compressed on Windows
mimetypes.add_type("application/zip", ".zip")


class Config():
    """Config class whose values are lazily-initialized when accessed.

    Values are loaded from config files relative to CWD. Consider calling
    load() if the script changes CWD during the runtime.
    """
    DEFAULT = {
        'app': {
            'name': 'WebScrapBook',
            'theme': 'default',
            'root': '.',
            'base': '',
            'index': '',
            'backup_dir': '.wsb/backup',
            'content_security_policy': 'strict',
            'allowed_x_for': '0',
            'allowed_x_proto': '0',
            'allowed_x_host': '0',
            'allowed_x_port': '0',
            'allowed_x_prefix': '0',
            },
        'server': {
            'port': '8080',
            'host': 'localhost',
            'ssl_on': 'false',
            'ssl_key': '',
            'ssl_cert': '',
            'ssl_pw': '',
            'browse': 'true',
            },
        'browser': {
            'command': '',
            'cache_prefix': 'webscrapbook.',
            'cache_expire': '259200',
            'use_jar': 'false',
            },
        'book ""': {
            'name': 'scrapbook',
            'top_dir': '',
            'data_dir': '',
            'tree_dir': '.wsb/tree',
            'index': '.wsb/tree/map.html',
            'no_tree': 'false',
            },
        }
    TYPES = {
        'app': {
            'allowed_x_for': 'getint',
            'allowed_x_proto': 'getint',
            'allowed_x_host': 'getint',
            'allowed_x_port': 'getint',
            'allowed_x_prefix': 'getint',
            },
        'server': {
            'port': 'getint',
            'ssl_on': 'getboolean',
            'browse': 'getboolean',
            },
        'browser': {
            'cache_expire': 'getint',
            'use_jar': 'getboolean',
            },
        'book': {
            None: {
                'no_tree': 'getboolean',
                },
            },
        }
    SUBSECTED = ['book', 'auth']

    def __init__(self):
        self._conf = None
        self._data = None


    def __getitem__(self, key):
        if self._conf is None: self.load()  # lazy load
        return self._data[key]


    def __iter__(self):
        if self._conf is None: self.load()  # lazy load
        return iter(self._data)


    def getname(self, name):
        if self._conf is None: self.load()  # lazy load
        parts = name.split('.')
        if len(parts) == 3:
            sec, subsec, key = parts
            try:
                return self._conf[f'{sec} "{subsec}"'][key]
            except KeyError:
                pass
        elif len(parts) == 2:
            sec, key = parts
            try:
                return self._conf[sec][key]
            except KeyError:
                pass
        return None


    def dump(self, fh):
        if self._conf is None: self.load()  # lazy load
        self._conf.write(fh)


    def dump_object(self):
        """Dump configs as an object, with type casting.
        """
        if self._conf is None: self.load()  # lazy load
        return deepcopy(self._data)


    def load(self, root='.'):
        """Loads config files related to the given root directory.

        Skip if config file doesn't exist, but raise if not loadable.

        project config > user config > default config
        """
        def load_config(file):
            if not os.path.isfile(file):
                return

            try:
                parser = ConfigParser(interpolation=None)
                parser.read(file, encoding='UTF-8')
            except Exception:
                print(f'Error: Unable to load config from "{file}".', file=sys.stderr)
                raise
            else:
                for section in parser.sections():
                    # Handle subsected sections formatted as [section "subsection"].
                    # Also normalize [section] and [section  ""  ] to [section ""].
                    m = re.search(r'^(\S*)(?:\s*"([^"\]]*)"\s*)?$', section)
                    if m:
                        sec, subsec = m.group(1), m.group(2) or ''
                        if sec in self.SUBSECTED:
                            newsection = f'{sec} "{subsec}"'
                            conf.setdefault(newsection, self.DEFAULT.get(f'{sec} ""', OrderedDict()))
                            conf[newsection].update(parser[section])
                            continue

                    # conf.setdefault(...).update(...) doesn't work here as the
                    # setdefault may return the default value rather then a
                    # Section object.
                    conf.setdefault(section, OrderedDict())
                    conf[section].update(parser[section])

        # default config
        self._conf = conf = ConfigParser(interpolation=None)
        conf.read_dict(self.DEFAULT)

        # user config
        load_config(os.path.join(WSB_USER_DIR, WSB_CONFIG))
        load_config(WSB_USER_CONFIG)

        # book config
        load_config(os.path.join(root, WSB_DIR, WSB_CONFIG))

        # map subsections
        self._data = OrderedDict()
        for section in conf.sections():
            sectionobj = OrderedDict()
            m = re.search(r'^(\S*)(?:\s*"([^"\]]*)"\s*)?$', section)
            if m:
                sec, subsec = m.group(1), m.group(2) or ''
                if sec in self.SUBSECTED:
                    self._data.setdefault(sec, OrderedDict())[subsec] = sectionobj
                    for key in conf[section]:
                        try:
                            sectionobj[key] = getattr(conf[section], self.TYPES[sec][None][key])(key)
                        except KeyError:
                            sectionobj[key] = conf[section][key]
                    continue
            self._data[section] = sectionobj
            for key in conf[section]:
                try:
                    sectionobj[key] = getattr(conf[section], self.TYPES[section][key])(key)
                except KeyError:
                    sectionobj[key] = conf[section][key]

config = Config()



from unittest import mock
import unittest
import sys
import os
import shutil
import io
import zipfile
import re
from urllib.request import pathname2url
from tempfile import gettempdir
from webscrapbook import WSB_DIR, WSB_CONFIG
from webscrapbook import Config
from webscrapbook import cli

root_dir = os.path.abspath(os.path.dirname(__file__))
test_dir = os.path.join(root_dir, 'test_cli')
resource_dir = os.path.join(root_dir, '..', 'webscrapbook', 'resources')

class TestServe(unittest.TestCase):
    @mock.patch('sys.stdout', io.StringIO)
    @mock.patch('webscrapbook.cli.server.serve', autospec=True)
    def test_call(self, mock_serve):
        cli.cmd_serve({
            'root': test_dir,
            'browse': False,
            })

        mock_serve.assert_called_once_with(root=test_dir, browse=False)

class TestConfig(unittest.TestCase):
    def tearDown(self):
        try:
            shutil.rmtree(os.path.join(test_dir, WSB_DIR))
        except NotADirectoryError:
            os.remove(os.path.join(test_dir, WSB_DIR))
        except FileNotFoundError:
            pass

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_book_create(self, mock_stdout):
        cli.cmd_config({
            'root': test_dir,
            'book': True,
            'user': False,
            'all': False,
            'edit': False,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'config.ini')) as f1:
            with open(os.path.join(resource_dir, 'config.ini')) as f2:
                self.assertTrue(f1.read(), f2.read())

        self.assertNotEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_book_create2(self, mock_stdout):
        """No overwrite."""
        os.makedirs(os.path.join(test_dir, WSB_DIR), exist_ok=True)
        with open(os.path.join(test_dir, WSB_DIR, 'config.ini'), 'w') as f:
            f.write('dummy')

        cli.cmd_config({
            'root': test_dir,
            'book': True,
            'user': False,
            'all': False,
            'edit': False,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'config.ini')) as f1:
            self.assertTrue(f1.read(), 'dummy')

        self.assertEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.util.launch', autospec=True)
    def test_book_edit(self, mock_launch, mock_stdout):
        cli.cmd_config({
            'root': test_dir,
            'book': True,
            'user': False,
            'all': False,
            'edit': True,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'config.ini')) as f1:
            with open(os.path.join(resource_dir, 'config.ini')) as f2:
                self.assertTrue(f1.read(), f2.read())

        mock_launch.assert_called_once_with(os.path.join(test_dir, WSB_DIR, 'config.ini'))

        self.assertNotEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.util.launch', autospec=True)
    def test_book_edit2(self, mock_launch, mock_stdout):
        os.makedirs(os.path.join(test_dir, WSB_DIR), exist_ok=True)
        with open(os.path.join(test_dir, WSB_DIR, 'config.ini'), 'w') as f:
            f.write('foo')

        cli.cmd_config({
            'root': test_dir,
            'book': True,
            'user': False,
            'all': False,
            'edit': True,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'config.ini')) as f1:
            self.assertTrue(f1.read(), 'foo')

        mock_launch.assert_called_once_with(os.path.join(test_dir, WSB_DIR, 'config.ini'))

        self.assertEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_book_all(self, mock_stdout):
        cli.cmd_config({
            'root': test_dir,
            'book': True,
            'user': False,
            'all': True,
            'edit': False,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'config.ini')) as f1:
            with open(os.path.join(resource_dir, 'config.ini')) as f2:
                self.assertTrue(f1.read(), f2.read())

        with open(os.path.join(test_dir, WSB_DIR, 'serve.py')) as f1:
            with open(os.path.join(resource_dir, 'serve.py')) as f2:
                self.assertTrue(f1.read(), f2.read())

        with open(os.path.join(test_dir, WSB_DIR, 'app.py')) as f1:
            with open(os.path.join(resource_dir, 'app.py')) as f2:
                self.assertTrue(f1.read(), f2.read())

        self.assertNotEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_book_all2(self, mock_stdout):
        """No overwrite."""
        os.makedirs(os.path.join(test_dir, WSB_DIR), exist_ok=True)
        with open(os.path.join(test_dir, WSB_DIR, 'config.ini'), 'w') as f:
            f.write('dummy1')
        with open(os.path.join(test_dir, WSB_DIR, 'serve.py'), 'w') as f:
            f.write('dummy2')
        with open(os.path.join(test_dir, WSB_DIR, 'app.py'), 'w') as f:
            f.write('dummy3')

        cli.cmd_config({
            'root': test_dir,
            'book': True,
            'user': False,
            'all': True,
            'edit': False,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'config.ini')) as f1:
            self.assertTrue(f1.read(), 'dummy1')

        with open(os.path.join(test_dir, WSB_DIR, 'serve.py')) as f1:
            self.assertTrue(f1.read(), 'dummy2')

        with open(os.path.join(test_dir, WSB_DIR, 'app.py')) as f1:
            self.assertTrue(f1.read(), 'dummy3')

        self.assertEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.cli.WSB_USER_CONFIG', os.path.join(test_dir, WSB_DIR, 'userconfig.ini'))
    def test_user_create(self, mock_stdout):
        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': True,
            'all': False,
            'edit': False,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'userconfig.ini')) as f1:
            with open(os.path.join(resource_dir, 'config.ini')) as f2:
                self.assertTrue(f1.read(), f2.read())

        self.assertNotEqual(mock_stdout.getvalue(), '')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.cli.WSB_USER_CONFIG', os.path.join(test_dir, WSB_DIR, 'userconfig.ini'))
    def test_user_create2(self, mock_stdout):
        os.makedirs(os.path.join(test_dir, WSB_DIR), exist_ok=True)
        with open(os.path.join(test_dir, WSB_DIR, 'userconfig.ini'), 'w') as f:
            f.write('dummy')

        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': True,
            'all': False,
            'edit': False,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'userconfig.ini')) as f1:
            self.assertTrue(f1.read(), 'dummy')

        self.assertEqual(mock_stdout.getvalue(), '')

    @mock.patch('webscrapbook.cli.WSB_USER_CONFIG', os.path.join(test_dir, WSB_DIR, 'userconfig.ini'))
    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.util.launch', autospec=True)
    def test_user_edit(self, mock_launch, mock_stdout):
        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': True,
            'all': False,
            'edit': True,
            'name': None,
            })

        with open(os.path.join(test_dir, WSB_DIR, 'userconfig.ini')) as f1:
            with open(os.path.join(resource_dir, 'config.ini')) as f2:
                self.assertTrue(f1.read(), f2.read())

        mock_launch.assert_called_once_with(os.path.join(test_dir, WSB_DIR, 'userconfig.ini'))

        self.assertNotEqual(mock_stdout.getvalue(), '')

    @mock.patch('webscrapbook.cli.WSB_USER_CONFIG', os.path.join(test_dir, WSB_DIR, 'userconfig.ini'))
    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.util.launch', autospec=True)
    def test_user_edit2(self, mock_launch, mock_stdout):
        os.makedirs(os.path.join(test_dir, WSB_DIR), exist_ok=True)
        with open(os.path.join(test_dir, WSB_DIR, 'userconfig.ini'), 'w') as f:
            f.write('dummy')

        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': True,
            'all': False,
            'edit': True,
            'name': None,
            })

        mock_launch.assert_called_once_with(os.path.join(test_dir, WSB_DIR, 'userconfig.ini'))

        self.assertEqual(mock_stdout.getvalue(), '')

    @mock.patch('webscrapbook.cli.die', side_effect=SystemExit)
    def test_edit(self, mock_die):
        with self.assertRaises(SystemExit):
            cli.cmd_config({
                'root': test_dir,
                'book': False,
                'user': False,
                'all': False,
                'edit': True,
                'name': None,
                })

        mock_die.assert_called_once_with('Use --edit in combine with --book or --user.')

    @mock.patch('webscrapbook.cli.die', side_effect=SystemExit)
    def test_all(self, mock_die):
        with self.assertRaises(SystemExit):
            cli.cmd_config({
                'root': test_dir,
                'book': False,
                'user': False,
                'all': True,
                'edit': False,
                'name': None,
                })

        mock_die.assert_called_once_with('Use --all in combine with --book.')

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.config.getname', return_value='dummy', autospec=True)
    def test_name_call(self, mock_getname, mock_stdout):
        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': False,
            'all': False,
            'edit': False,
            'name': 'app.name',
            })

        mock_getname.assert_called_once_with('app.name')
        self.assertEqual(mock_stdout.getvalue(), 'dummy\n')

    @mock.patch('webscrapbook.cli.die', side_effect=SystemExit)
    @mock.patch('webscrapbook.config.getname', return_value=None, autospec=True)
    def test_name_call2(self, mock_getname, mock_die):
        with self.assertRaises(SystemExit):
            cli.cmd_config({
                'root': test_dir,
                'book': False,
                'user': False,
                'all': False,
                'edit': False,
                'name': 'unknown.config',
                })

        mock_getname.assert_called_once_with('unknown.config')
        mock_die.assert_called_once_with('Config entry "unknown.config" does not exist')

    @mock.patch('webscrapbook.config.dump', autospec=True)
    def test_dump(self, mock_dump):
        cli.cmd_config({
            'root': test_dir,
            'book': False,
            'user': False,
            'all': False,
            'edit': False,
            'name': None,
            })

        mock_dump.assert_called_once_with(sys.stdout)

class TestEncrypt(unittest.TestCase):
    @mock.patch('sys.stdout', new_callable=io.StringIO)
    @mock.patch('webscrapbook.util.encrypt', return_value='dummy_hash', autospec=True)
    def test_call(self, mock_encrypt, mock_stdout):
        cli.cmd_encrypt({
            'password': '1234',
            'salt': 'mysalt',
            'method': 'sha256',
            })

        mock_encrypt.assert_called_once_with('1234', salt='mysalt', method='sha256')
        self.assertEqual(mock_stdout.getvalue(), 'dummy_hash\n')

class TestCache(unittest.TestCase):
    @mock.patch('webscrapbook.scrapbook.cache.generate', autospec=True)
    def test_call(self, mock_func):
        cli.cmd_cache({
            'root': test_dir,
            'book_ids': ['book1', 'book2'],
            'item_ids': ['item1', 'item2'],
            'fulltext': False,
            'inclusive_frames': False,
            'static_site': True,
            'static_index': True,
            'rss_root': 'http://example.com:8000/wsb/',
            'locale': 'zh_TW',
            'no_backup': True,
            'debug': True,
            })

        mock_func.assert_called_once_with(
            root=test_dir,
            book_ids=['book1', 'book2'],
            item_ids=['item1', 'item2'],
            fulltext=False,
            inclusive_frames=False,
            static_site=True,
            static_index=True,
            rss_root='http://example.com:8000/wsb/',
            locale='zh_TW',
            no_backup=True,
            )

class TestCheck(unittest.TestCase):
    @mock.patch('webscrapbook.scrapbook.check.run', autospec=True)
    def test_call(self, mock_func):
        cli.cmd_check({
            'root': test_dir,
            'book_ids': ['book1', 'book2'],
            'resolve_all': False,
            'resolve_invalid_id': True,
            'resolve_missing_index': True,
            'resolve_missing_index_file': True,
            'resolve_missing_date': True,
            'resolve_older_mtime': True,
            'resolve_toc_unreachable': True,
            'resolve_toc_invalid': True,
            'resolve_toc_empty_subtree': True,
            'resolve_unindexed_files': True,
            'resolve_absolute_icon': True,
            'resolve_unused_icon': True,
            'no_backup': True,
            'debug': True,
            })

        mock_func.assert_called_once_with(
            root=test_dir,
            book_ids=['book1', 'book2'],
            resolve_all=False,
            resolve_invalid_id=True,
            resolve_missing_index=True,
            resolve_missing_index_file=True,
            resolve_missing_date=True,
            resolve_older_mtime=True,
            resolve_toc_unreachable=True,
            resolve_toc_invalid=True,
            resolve_toc_empty_subtree=True,
            resolve_unindexed_files=True,
            resolve_absolute_icon=True,
            resolve_unused_icon=True,
            no_backup=True,
            )

class TestExport(unittest.TestCase):
    @mock.patch('webscrapbook.scrapbook.exporter.run', autospec=True)
    def test_call(self, mock_func):
        cli.cmd_export({
            'root': test_dir,
            'output': os.path.join(test_dir, 'export'),
            'book_id': 'id1',
            'item_ids': ['item1', 'item2'],
            'recursive': True,
            'singleton': True,
            'debug': True,
            })

        mock_func.assert_called_once_with(
            root=test_dir,
            output=os.path.join(test_dir, 'export'),
            book_id='id1',
            item_ids=['item1', 'item2'],
            recursive=True,
            singleton=True,
            )

class TestImport(unittest.TestCase):
    @mock.patch('webscrapbook.scrapbook.importer.run', autospec=True)
    def test_call(self, mock_func):
        cli.cmd_import({
            'root': test_dir,
            'files': [os.path.join(test_dir, 'import')],
            'book': 'id1',
            'target_id': '20200101000000000',
            'target_index': 1,
            'target_filename': '%UUID%',
            'rebuild_folders': True,
            'resolve_id_used': 'new',
            'prune': True,
            'debug': True,
            })

        mock_func.assert_called_once_with(
            root=test_dir,
            files=[os.path.join(test_dir, 'import')],
            book='id1',
            target_id='20200101000000000',
            target_index=1,
            target_filename='%UUID%',
            rebuild_folders=True,
            resolve_id_used='new',
            prune=True,
            )

class TestConvert(unittest.TestCase):
    @mock.patch('webscrapbook.scrapbook.convert.sb2wsb.run', autospec=True)
    def test_sb2wsb(self, mock_func):
        cli.cmd_convert({
            'root': test_dir,
            'mode': 'sb2wsb',
            'input': test_dir,
            'output': os.path.join(root_dir, 'test_cli_output'),
            'no_backup': True,
            'force': True,
            'debug': True,
            })

        mock_func.assert_called_once_with(
            input=test_dir,
            output=os.path.join(root_dir, 'test_cli_output'),
            no_backup=True,
            )

    @mock.patch('webscrapbook.scrapbook.convert.wsb2sb.run', autospec=True)
    def test_wsb2sb(self, mock_func):
        cli.cmd_convert({
            'root': test_dir,
            'mode': 'wsb2sb',
            'input': test_dir,
            'output': os.path.join(root_dir, 'test_cli_output'),
            'book_id': 'id1',
            'force': True,
            'debug': True,
            })

        mock_func.assert_called_once_with(
            input=test_dir,
            output=os.path.join(root_dir, 'test_cli_output'),
            book_id='id1',
            )

class TestHelp(unittest.TestCase):
    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_call(self, mock_stdout):
        cli.cmd_help({
            'topic': 'config',
            })

        with open(os.path.join(resource_dir, 'config.md')) as f:
            self.assertEqual(mock_stdout.getvalue(), f.read() + '\n')

class TestView(unittest.TestCase):
    @mock.patch('webscrapbook.cli.view_archive_files', autospec=True)
    def test_call(self, mock_view):
        cli.cmd_view({
            'root': 'config',
            'files': [test_dir],
            })

        mock_view.assert_called_once_with([test_dir])

class TestHelpers(unittest.TestCase):
    def tearDown(self):
        try:
            shutil.rmtree(os.path.join(test_dir, 'temp'))
        except NotADirectoryError:
            os.remove(os.path.join(test_dir, 'temp'))
        except FileNotFoundError:
            pass

    @mock.patch('tempfile.gettempdir', lambda: os.path.join(test_dir, 'temp', 'tempdir'))
    @mock.patch('webbrowser.get', new_callable=mock.MagicMock())
    def test_view_archive_files(self, mock_browser):
        testfile1 = os.path.join(test_dir, 'temp', 'test1.htz')
        testfile2 = os.path.join(test_dir, 'temp', 'test2.maff')
        testfile3 = os.path.join(test_dir, 'temp', 'test3.maff')
        testfile4 = os.path.join(test_dir, 'temp', 'test4.zip')
        testtemp = os.path.join(test_dir, 'temp', 'tempdir')
        testtempprefix = os.path.join(test_dir, 'temp', 'tempdir', 'webscrapbook.')

        os.makedirs(testtemp, exist_ok=True)
        with zipfile.ZipFile(testfile1, 'w') as zh:
            zh.writestr('index.html', 'abc 測試')
        with zipfile.ZipFile(testfile2, 'w') as zh:
            zh.writestr('123456/index.html', 'abc 測試')
        with zipfile.ZipFile(testfile3, 'w') as zh:
            zh.writestr('abc/index.html', 'abc 測試')
            zh.writestr('def/index.html', 'abc 測試')
        with zipfile.ZipFile(testfile4, 'w') as zh:
            zh.writestr('index.html', 'abc 測試')

        # test jar, no temp dir should be generated
        with mock.patch('webscrapbook.cli.config', {
                'browser': {
                    'command': '',
                    'cache_prefix': 'webscrapbook.',
                    'cache_expire': 1000,
                    'use_jar': True,
                    }
                }):
            cli.view_archive_files([testfile1, testfile2, testfile3, testfile4])

        self.assertEqual(len(mock_browser.mock_calls), 5)
        self.assertEqual(mock_browser.mock_calls[1][1][0],
            fr'jar:file:{pathname2url(os.path.normcase(testfile1))}!/index.html')
        self.assertEqual(mock_browser.mock_calls[2][1][0],
            fr'jar:file:{pathname2url(os.path.normcase(testfile2))}!/123456/index.html')
        self.assertEqual(mock_browser.mock_calls[3][1][0],
            fr'jar:file:{pathname2url(os.path.normcase(testfile3))}!/abc/index.html')
        self.assertEqual(mock_browser.mock_calls[4][1][0],
            fr'jar:file:{pathname2url(os.path.normcase(testfile3))}!/def/index.html')
        self.assertEqual(len(os.listdir(testtemp)), 0)

        # test simple view
        mock_browser.reset_mock()
        with mock.patch('webscrapbook.cli.config', {
                'browser': {
                    'command': '',
                    'cache_prefix': 'webscrapbook.',
                    'cache_expire': 1000,
                    'use_jar': False,
                    }
                }):
            cli.view_archive_files([testfile1, testfile2, testfile3, testfile4])

        mock_browser.assert_called_once_with(None)
        self.assertEqual(len(mock_browser.mock_calls), 5)
        self.assertRegex(mock_browser.mock_calls[1][1][0],
            fr'^file:{re.escape(pathname2url(testtemp))}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/index\.html$')
        self.assertRegex(mock_browser.mock_calls[2][1][0],
            fr'^file:{re.escape(pathname2url(testtemp))}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/123456/index\.html$')
        self.assertRegex(mock_browser.mock_calls[3][1][0],
            fr'^file:{re.escape(pathname2url(testtemp))}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/abc/index\.html$')
        self.assertRegex(mock_browser.mock_calls[4][1][0],
            fr'^file:{re.escape(pathname2url(testtemp))}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/def/index\.html$')
        self.assertEqual(len(os.listdir(testtemp)), 3)

        # test browser command
        # test if cache is used for same archive
        mock_browser.reset_mock()
        with mock.patch('webscrapbook.cli.config', {
                'browser': {
                    'command': '/path/to/firefox',
                    'cache_prefix': 'webscrapbook.',
                    'cache_expire': 1000,
                    'use_jar': False,
                    }
                }):
            cli.view_archive_files([testfile1, testfile2, testfile3, testfile4])

        mock_browser.assert_called_once_with('/path/to/firefox')
        self.assertEqual(len(mock_browser.mock_calls), 5)
        self.assertRegex(mock_browser.mock_calls[1][1][0],
            fr'^file:{re.escape(pathname2url(testtemp))}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/index\.html$')
        self.assertRegex(mock_browser.mock_calls[2][1][0],
            fr'^file:{re.escape(pathname2url(testtemp))}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/123456/index\.html$')
        self.assertRegex(mock_browser.mock_calls[3][1][0],
            fr'^file:{re.escape(pathname2url(testtemp))}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/abc/index\.html$')
        self.assertRegex(mock_browser.mock_calls[4][1][0],
            fr'^file:{re.escape(pathname2url(testtemp))}/webscrapbook\.[0-9a-z]*_[0-9a-z_]*/def/index\.html$')
        self.assertEqual(len(os.listdir(testtemp)), 3)

        # test auto clearance of stale caches
        mock_browser.reset_mock()
        with mock.patch('webscrapbook.cli.config', {
                'browser': {
                    'command': '',
                    'cache_prefix': 'webscrapbook.',
                    'cache_expire': -1,
                    'use_jar': False,
                    }
                }):
            with mock.patch('shutil.rmtree') as mock_rmtree:
                cli.view_archive_files([])
                self.assertEqual(len(mock_rmtree.call_args_list), 3)
                self.assertRegex(mock_rmtree.call_args_list[0][0][0].path,
                    fr'^{re.escape(testtempprefix)}[0-9a-z]*_[0-9a-z_]*$')
                self.assertRegex(mock_rmtree.call_args_list[1][0][0].path,
                    fr'^{re.escape(testtempprefix)}[0-9a-z]*_[0-9a-z_]*$')
                self.assertRegex(mock_rmtree.call_args_list[2][0][0].path,
                    fr'^{re.escape(testtempprefix)}[0-9a-z]*_[0-9a-z_]*$')

if __name__ == '__main__':
    unittest.main()



# since Python 3.6
try:
    from contextlib import AbstractContextManager
except ImportError:
    import abc
    import _collections_abc

    class AbstractContextManager(abc.ABC):

        """An abstract base class for context managers."""

        def __enter__(self):
            """Return `self` upon entering the runtime context."""
            return self

        @abc.abstractmethod
        def __exit__(self, exc_type, exc_value, traceback):
            """Raise any exception triggered within the runtime context."""
            return None

        @classmethod
        def __subclasshook__(cls, C):
            if cls is AbstractContextManager:
                return _collections_abc._check_methods(C, "__enter__", "__exit__")
            return NotImplemented


# since Python 3.7
try:
    from contextlib import nullcontext
except ImportError:
    class nullcontext(AbstractContextManager):
        """Context manager that does no additional processing.
        Used as a stand-in for a normal context manager, when a particular
        block of code is only sometimes used with a normal context manager:
        cm = optional_cm if condition else nullcontext()
        with cm:
            # Perform operation, using optional_cm if condition is True
        """

        def __init__(self, enter_result=None):
            self.enter_result = enter_result

        def __enter__(self):
            return self.enter_result

        def __exit__(self, *excinfo):
            pass



import os
import traceback
import zipfile
import time
import json
from datetime import datetime, timezone, timedelta

from .host import Host
from .. import util
from ..util import Info
from .._compat.contextlib import nullcontext


class Exporter():
    """Main class for generating exports.
    """
    def __init__(self, output, book, *, singleton=False):
        self.output = output
        self.book = book
        self.singleton = singleton

        self.recycle = None
        self.used_ts = None
        self.map_id_to_eid = None

    def run(self, item_ids=None, recursive=False):
        book = self.book
        self.book.load_meta_files()
        self.book.load_toc_files()

        os.makedirs(self.output, exist_ok=True)

        self.recycle = set(book.toc.get('recycle', []))
        self.used_ts = set()
        self.map_id_to_eid = {}

        if item_ids:
            id_pool = set(id for id in item_ids if id in book.meta and id not in self.recycle)

            # add descendant id if recursive mode
            if recursive:
                for id in list(id_pool):
                    for desc_id in self._iter_child_items(id, [id]):
                        id_pool.add(desc_id)
                        yield Info('debug', f'Included descendant item of "{id}": "{desc_id}"')
        else:
            id_pool = set(id for id in book.meta if id not in self.recycle)

        id_chain = ['root']
        for id in self._iter_child_items('root', id_chain):
            if id in id_pool:
                yield from self._export_item(id, id_chain)

        id_chain = ['hidden']
        for id in self._iter_child_items('hidden', id_chain):
            if id in id_pool:
                yield from self._export_item(id, id_chain)

    def _iter_child_items(self, id, id_chain):
        """Generate descendant items for the item of id.
        """
        for ref_id in self.book.toc.get(id, []):
            yield ref_id

            # do not export children of a circular item
            if ref_id not in id_chain:
                id_chain.append(ref_id)
                yield from self._iter_child_items(ref_id, id_chain)
                id_chain.pop()

    def _export_item(self, id, id_chain):
        if id in self.map_id_to_eid:
            if self.singleton:
                yield Info('debug', f'Skipped exporting item "{id}" (singleton mode)')
                return

        yield Info('debug', f'Exporting item "{id}"')
        try:
            yield from self._export_item_internal(id, id_chain)
        except Exception as exc:
            # unexpected error
            traceback.print_exc()
            yield Info('error', f'Failed to export "{id}": {exc}', exc=exc)

    def _export_item_internal(self, id, id_chain):
        meta = self.book.meta[id]
        index = meta.get('index', '')

        # generate a unique timestamp as prefix
        ts = datetime.now(timezone.utc)
        while ts in self.used_ts:
            ts += timedelta(milliseconds=1)
        self.used_ts.add(ts)
        ets = util.datetime_to_id(ts)

        # setup an export id (eid), which is unique and is same among all
        # occurrences of the same id, to the ets of the first occurrence
        eid = self.map_id_to_eid.setdefault(id, ets)

        # generate a unique timestamp prefix
        basename = ets + '-' + meta.get('title', meta.get('id', ''))
        basename = util.crop(util.validate_filename(basename), 128)

        # generate a unique filename
        i = 0
        dst = os.path.join(self.output, f'{basename}.wsba')
        while os.path.lexists(dst):
            i += 1
            dst = os.path.join(self.output, f'{basename}-{i}.wsba')

        yield Info('info', f'Exporting "{id}" to "{os.path.basename(dst)}"')
        parents = [{'id': id, 'title': self.book.meta.get(id, {}).get('title', '')} for id in id_chain]
        meta_data = {'id': id, **meta}
        export_data = {
            'version': 1,
            'id': eid,
            'timestamp': ets,
            'timezone': datetime.now().astimezone().utcoffset().total_seconds(),
            'path': parents,
            }
        with zipfile.ZipFile(dst, 'w') as zh:
            zh.writestr('meta.json', json.dumps(meta_data, ensure_ascii=False, indent=2))
            zh.writestr('export.json', json.dumps(export_data, ensure_ascii=False, indent=2))

            # include data file(s)
            if index:
                zh.writestr('data/', '')
                src = os.path.join(self.book.data_dir, os.path.dirname(index) if index.endswith('/index.html') else index)
                yield Info('debug', f'Saving data files for "{id}": "{self.book.get_subpath(src)}"')
                util.zip_compress(zh, src, f'data/{os.path.basename(src)}')

            # include favicon cache
            iconfile = self.book.get_icon_file(meta)
            if not iconfile:
                return

            favicon_dir = os.path.join(self.book.tree_dir, 'favicon', '')
            if not os.path.normcase(iconfile).startswith(os.path.normcase(favicon_dir)):
                return

            zh.writestr('favicon/', '')
            util.zip_compress(zh, iconfile, f'favicon/{os.path.basename(iconfile)}')


def run(root, output, book_id='', item_ids=None, *, recursive=False, singleton=False,
        config=None, no_lock=False):
    start = time.time()

    host = Host(root, config)

    # Fail for invalid book ID
    if book_id not in host.books:
        yield Info('error', f'Invalid book "{book_id}".')
        return

    yield Info('debug', f'Loading book "{book_id}".')

    try:
        book = host.books[book_id]

        if book.no_tree:
            yield Info('error', f'Unable to export book "{book_id}" ("{book.name}") (no_tree).')
            return

        yield Info('info', f'Exporting from book "{book_id}" ({book.name}).')
        lh = nullcontext() if no_lock else book.get_tree_lock().acquire()
        with lh:
            generator = Exporter(output, book, singleton=singleton)
            yield from generator.run(item_ids, recursive)

    except Exception as exc:
        traceback.print_exc()
        yield Info('critical', str(exc), exc=exc)
    else:
        yield Info('info', 'Done.')

    yield Info('info', '----------------------------------------------------------------------')

    elapsed = time.time() - start
    yield Info('info', f'Time spent: {elapsed} seconds.')



from unittest import mock
import unittest
import os
import shutil
import time
from webscrapbook import WSB_DIR
from webscrapbook import util
from webscrapbook.scrapbook.host import Host
from webscrapbook.scrapbook.convert import sb2wsb

root_dir = os.path.abspath(os.path.dirname(__file__))
test_root = os.path.join(root_dir, 'test_scrapbook_convert')

def setUpModule():
    # mock out user config
    global mockings
    mockings = [
        mock.patch('webscrapbook.scrapbook.host.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_DIR', os.path.join(test_root, 'wsb')),
        mock.patch('webscrapbook.WSB_USER_CONFIG', test_root),
        ]
    for mocking in mockings:
        mocking.start()

def tearDownModule():
    # stop mock
    for mocking in mockings:
        mocking.stop()

class TestRun(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = 8192
        cls.test_input = os.path.join(test_root, 'input')
        cls.test_input_rdf = os.path.join(cls.test_input, 'scrapbook.rdf')
        cls.test_output = os.path.join(test_root, 'output')
        cls.test_output_tree = os.path.join(cls.test_output, 'tree')
        cls.test_output_meta = os.path.join(cls.test_output_tree, 'meta.js')
        cls.test_output_toc = os.path.join(cls.test_output_tree, 'toc.js')

    def setUp(self):
        """Set up a general temp test folder
        """
        os.makedirs(self.test_input, exist_ok=True)
        os.makedirs(self.test_output, exist_ok=True)

    def tearDown(self):
        """Remove general temp test folder
        """
        try:
            shutil.rmtree(self.test_input)
        except NotADirectoryError:
            os.remove(self.test_input)
        except FileNotFoundError:
            pass

        try:
            shutil.rmtree(self.test_output)
        except NotADirectoryError:
            os.remove(self.test_output)
        except FileNotFoundError:
            pass

    def test_meta_basic01(self):
        """A typical item sample of legacy ScrapBook X."""
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200102030405"
                   NS1:id="20200102030405"
                   NS1:create="20200102030406"
                   NS1:modify="20200102030407"
                   NS1:type=""
                   NS1:title="dummy title"
                   NS1:chars="UTF-8"
                   NS1:icon="favicon.ico"
                   NS1:source="http://example.com/foo"
                   NS1:comment="dummy comment __BR__ 2nd line"
                   NS1:lock="true" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertDictEqual(book.meta['20200102030405'], {
            'index': '20200102030405/index.html',
            'title': 'dummy title',
            'type': '',
            'create': util.datetime_to_id(util.id_to_datetime_legacy('20200102030406')),
            'modify': util.datetime_to_id(util.id_to_datetime_legacy('20200102030407')),
            'source': 'http://example.com/foo',
            'icon': 'favicon.ico',
            'comment': 'dummy comment\n2nd line',
            'charset': 'UTF-8',
            'locked': True,
            })

    def test_meta_basic02(self):
        """A typical item sample of legacy ScrapBook."""
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200102030405"
                   NS1:id="20200102030405"
                   NS1:type=""
                   NS1:title="dummy title"
                   NS1:chars=""
                   NS1:comment="dummy comment __BR__ 2nd line"
                   NS1:icon="favicon.ico"
                   NS1:source="http://example.com/foo" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertDictEqual(book.meta['20200102030405'], {
            'index': '20200102030405/index.html',
            'title': 'dummy title',
            'type': '',
            'create': util.datetime_to_id(util.id_to_datetime_legacy('20200102030405')),
            'modify': util.datetime_to_id(util.id_to_datetime_legacy('20200102030405')),
            'source': 'http://example.com/foo',
            'icon': 'favicon.ico',
            'comment': 'dummy comment\n2nd line',
            })

    def test_meta_basic03(self):
        """Default value for missing keys.

        - type: ""
        - create: infer from id
        - modify: infer from create (and then id)
        """
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200102030405" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertDictEqual(book.meta['20200102030405'], {
            'index': '20200102030405/index.html',
            'type': '',
            'create': util.datetime_to_id(util.id_to_datetime_legacy('20200102030405')),
            'modify': util.datetime_to_id(util.id_to_datetime_legacy('20200102030405')),
            })

    def test_meta_basic04(self):
        """Should work correctly for another NS name

        (seen in some scrapbook.rdf files)
        """
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS2="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200102030405"
                   NS2:id="20200102030405"
                   NS2:type=""
                   NS2:title="title"
                   NS2:comment="comment"
                   NS2:icon="favicon.ico"
                   NS2:source="http://example.com" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertDictEqual(book.meta['20200102030405'], {
            'index': '20200102030405/index.html',
            'type': '',
            'create': util.datetime_to_id(util.id_to_datetime_legacy('20200102030405')),
            'modify': util.datetime_to_id(util.id_to_datetime_legacy('20200102030405')),
            'title': 'title',
            'comment': 'comment',
            'icon': 'favicon.ico',
            'source': 'http://example.com',
            })

    def test_meta_type01(self):
        """Translate types."""
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200101010101"
                   NS1:type="" />
  <RDF:Description RDF:about="urn:scrapbook:item20200101010102"
                   NS1:type="combine" />
  <RDF:Description RDF:about="urn:scrapbook:item20200101010103"
                   NS1:type="marked" />
  <RDF:Description RDF:about="urn:scrapbook:item20200101010104"
                   NS1:type="note" />
  <RDF:Description RDF:about="urn:scrapbook:item20200101010105"
                   NS1:type="notex" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertEqual(book.meta['20200101010101']['type'], '')
        self.assertIsNone(book.meta['20200101010101'].get('marked'))
        self.assertEqual(book.meta['20200101010102']['type'], 'combine')
        self.assertIsNone(book.meta['20200101010102'].get('marked'))
        self.assertEqual(book.meta['20200101010103']['type'], '')
        self.assertTrue(book.meta['20200101010103']['marked'])
        self.assertEqual(book.meta['20200101010104']['type'], 'postit')
        self.assertIsNone(book.meta['20200101010104'].get('marked'))
        self.assertEqual(book.meta['20200101010105']['type'], 'note')
        self.assertIsNone(book.meta['20200101010105'].get('marked'))

    def test_meta_type02(self):
        """Use "site" if it's "marked" and has sitemap.xml"""
        index_file = os.path.join(self.test_input, 'data', '20200101010101', 'index.html')
        sitemap_file = os.path.join(self.test_input, 'data', '20200101010101', 'sitemap.xml')
        os.makedirs(os.path.dirname(sitemap_file))
        with open(index_file, 'w', encoding='UTF-8') as fh:
            pass
        with open(sitemap_file, 'w', encoding='UTF-8') as fh:
            pass

        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200101010101"
                   NS1:type="marked" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertEqual(book.meta['20200101010101']['type'], 'site')
        self.assertTrue(book.meta['20200101010101']['marked'])

    def test_meta_modify01(self):
        """Take mtime of index.html if newer."""
        index_file = os.path.join(self.test_input, 'data', '20200102030405', 'index.html')
        os.makedirs(os.path.dirname(index_file))
        with open(index_file, 'w', encoding='UTF-8') as fh:
            pass
        t = time.mktime((2020, 1, 3, 0, 0, 0, 0, 0, -1))
        os.utime(index_file, (t, t))
        
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200102030405"
                   NS1:id="20200102030405"
                   NS1:type=""
                   NS1:create="20200102030406"
                   NS1:modify="20200102030407" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertEqual(book.meta['20200102030405']['modify'],
            util.datetime_to_id(util.id_to_datetime_legacy('20200103000000')))

    def test_meta_modify02(self):
        """Don't take mtime of index.html if not newer."""
        index_file = os.path.join(self.test_input, 'data', '20200102030405', 'index.html')
        os.makedirs(os.path.dirname(index_file))
        with open(index_file, 'w', encoding='UTF-8') as fh:
            pass
        t = time.mktime((2020, 1, 1, 0, 0, 0, 0, 0, -1))
        os.utime(index_file, (t, t))
        
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200102030405"
                   NS1:id="20200102030405"
                   NS1:type=""
                   NS1:create="20200102030406"
                   NS1:modify="20200102030407" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertEqual(book.meta['20200102030405']['modify'],
            util.datetime_to_id(util.id_to_datetime_legacy('20200102030407')))

    def test_meta_icon01(self):
        """Resolve resource://scrapbook/data/<self-id>/..."""
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200102030405"
                   NS1:icon="resource://scrapbook/data/20200102030405/favicon.ico" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertEqual(book.meta['20200102030405']['icon'], 'favicon.ico')

    def test_meta_icon02(self):
        """Resolve resource://scrapbook/icon/..."""
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200102030405"
                   NS1:icon="resource://scrapbook/icon/favicon.ico" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertEqual(book.meta['20200102030405']['icon'], '../../icon/favicon.ico')

    def test_meta_icon03(self):
        """Resolve resource://scrapbook/..."""
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200102030405"
                   NS1:icon="resource://scrapbook/favicon.ico" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertEqual(book.meta['20200102030405']['icon'], '../../favicon.ico')

    def test_meta_icon04(self):
        """Resolve moz-icon://..."""
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200102030405"
                   NS1:icon="moz-icon://foo.pdf?size=16" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertEqual(book.meta['20200102030405']['icon'], '')
        self.assertEqual(book.meta['20200102030405']['icon-moz'], 'moz-icon://foo.pdf?size=16')

    def test_meta_icon05(self):
        """Handle special chars."""
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Description RDF:about="urn:scrapbook:item20200102030405"
                   NS1:icon="resource://scrapbook/%E4%B8%AD%E6%96%87%25%23abc.png" />
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_meta_files()

        self.assertEqual(book.meta['20200102030405']['icon'], '../../%E4%B8%AD%E6%96%87%25%23abc.png')

    def test_toc(self):
        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
  <RDF:Seq RDF:about="urn:scrapbook:root">
    <RDF:li RDF:resource="urn:scrapbook:item20200101000000"/>
    <RDF:li RDF:resource="urn:scrapbook:item20200101000001"/>
    <RDF:li RDF:resource="urn:scrapbook:item20200101000002"/>
    <RDF:li RDF:resource="urn:scrapbook:item20200101000003"/>
    <RDF:li RDF:resource="urn:scrapbook:item20200101000004"/>
  </RDF:Seq>
  <RDF:Seq RDF:about="urn:scrapbook:item20200101000001">
    <RDF:li RDF:resource="urn:scrapbook:item20200101000005"/>
    <RDF:li RDF:resource="urn:scrapbook:item20200101000006"/>
  </RDF:Seq>
  <RDF:Seq RDF:about="urn:scrapbook:item20200101000002">
    <RDF:li RDF:resource="urn:scrapbook:item20200101000007"/>
    <RDF:li RDF:resource="urn:scrapbook:item20200101000008"/>
  </RDF:Seq>
  <RDF:Seq RDF:about="urn:scrapbook:item20200101000003">
    <RDF:li RDF:resource="urn:scrapbook:item20200101000009"/>
  </RDF:Seq>
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        book = Host(self.test_output).books['']
        book.load_toc_files()

        self.assertEqual(book.toc, {
            'root': [
                '20200101000000',
                '20200101000001',
                '20200101000002',
                '20200101000003',
                '20200101000004',
                ],
            '20200101000001': [
                '20200101000005',
                '20200101000006',
                ],
            '20200101000002': [
                '20200101000007',
                '20200101000008',
                ],
            '20200101000003': [
                '20200101000009',
                ],
            })

    def test_backup01(self):
        """Check legacy scrapbook files are backuped"""
        check_entries = [
            'backup/',
            'tree/',
            'scrapbook.rdf',
            'cache.rdf',
            'folders.txt',
            'collection.html',
            'combine.html',
            'combine.css',
            'note.html',
            'search.html',
            'sitemap.xsl',
            'note_template.html',
            'notex_template.html',
            ]

        for entry in check_entries:
            path = os.path.join(self.test_input, entry)
            if entry.endswith('/'):
                os.makedirs(path)
                with open(os.path.join(path, 'dummy'), 'w', encoding='UTF-8') as fh:
                    fh.write(entry)
            else:
                with open(path, 'w', encoding='UTF-8') as fh:
                    fh.write(entry)

        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output):
            pass

        with os.scandir(os.path.join(self.test_output, WSB_DIR, 'backup')) as dirs:
            backup_dir = next(iter(dirs))
        for entry in check_entries:
            with self.subTest(entry=entry):
                path = os.path.join(backup_dir, entry)
                if entry.endswith('/'):
                    self.assertTrue(os.path.isdir(path))
                    self.assertTrue(os.path.isfile(os.path.join(path, 'dummy')))
                else:
                    self.assertTrue(os.path.isfile(path))

    def test_backup02(self):
        """Check no_backup"""
        check_entries = [
            'backup/',
            'tree/',
            'scrapbook.rdf',
            'cache.rdf',
            'folders.txt',
            'collection.html',
            'combine.html',
            'combine.css',
            'note.html',
            'search.html',
            'sitemap.xsl',
            'note_template.html',
            'notex_template.html',
            ]

        for entry in check_entries:
            path = os.path.join(self.test_input, entry)
            if entry.endswith('/'):
                os.makedirs(path)
                with open(os.path.join(path, 'dummy'), 'w', encoding='UTF-8') as fh:
                    fh.write(entry)
            else:
                with open(path, 'w', encoding='UTF-8') as fh:
                    fh.write(entry)

        with open(self.test_input_rdf, 'w', encoding='UTF-8') as fh:
            fh.write("""\
<?xml version="1.0"?>
<RDF:RDF xmlns:NS1="http://amb.vis.ne.jp/mozilla/scrapbook-rdf#"
         xmlns:NC="http://home.netscape.com/NC-rdf#"
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
</RDF:RDF>
""")

        for info in sb2wsb.run(self.test_input, self.test_output, no_backup=True):
            pass

        self.assertFalse(os.path.exists(os.path.join(self.test_output, WSB_DIR, 'backup')))

if __name__ == '__main__':
    unittest.main()



"""Scrapbook book handler.
"""
import os
import zipfile
import re
import json
from urllib.parse import urlsplit, quote, unquote

from lxml import etree

from .. import WSB_DIR
from .. import util
from .._compat import zip_stream


class TreeFileError(ValueError):
    def __init__(self, msg, filename=None):
        self.msg = msg
        self.filename = filename


class TreeFileMalformedError(TreeFileError):
    pass


class TreeFileMalformedWrappingError(TreeFileMalformedError):
    """Malformed wrapping part of a tree file
    """


class TreeFileMalformedJsonError(TreeFileMalformedError):
    """A subclass for JSONDecodeError

    This exception should be raised from a JSONDecodeError, and details
        are accessible through __cause__ attribute.
    """


class Book:
    """Main scrapbook book controller.
    """
    REGEX_TREE_FILE_WRAPPER = re.compile(r'^(?:/\*.*\*/|[^(])+\(([\s\S]*)\)(?:/\*.*\*/|[\s;])*$')
    REGEX_ITEM_NOTE = re.compile(r'^.*?<pre>\n?([^<]*(?:<(?!/pre>)[^<]*)*)\n</pre>.*$', re.S)
    SAVE_META_THRESHOLD = 256 * 1024
    SAVE_TOC_THRESHOLD = 4 * 1024 * 1024
    SAVE_FULLTEXT_THRESHOLD = 128 * 1024 * 1024
    REPR_ATTRS = ('id', 'name', 'top_dir')
    DEFAULT_META = {
        'id': None,
        'index': None,
        'title': None,
        'type': None,
        'create': None,
        'modify': None,
        'source': None,
        'icon': None,
        'comment': None,
        }
    SPECIAL_ITEM_ID = {
        'root',
        'hidden',
        'recycle',
        }
    TYPES_OPTIONAL_INDEX = {
        'folder',
        'separator',
        'bookmark',
        }
    META_PROPERTY_PREFIX = 'data-scrapbook-'
    ITEM_NOTE_FORMATTER = """\
<!DOCTYPE html><html><head>\
<meta charset="UTF-8">\
<meta name="viewport" content="width=device-width">\
<style>pre { white-space: pre-wrap; overflow-wrap: break-word; }</style>\
</head><body><pre>
%NOTE_CONTENT%
</pre></body></html>"""

    def __init__(self, host, book_id=''):
        self.host = host
        config = host.config['book'][book_id]
        self.id = book_id
        self.name = config['name']
        self.root = host.root
        self.top_dir = os.path.normpath(os.path.join(host.chroot, config['top_dir']))
        self.data_dir = os.path.normpath(os.path.join(self.top_dir, config['data_dir']))
        self.tree_dir = os.path.normpath(os.path.join(self.top_dir, config['tree_dir']))
        self.no_tree = config['no_tree']

        self.meta = None
        self.toc = None
        self.fulltext = None
        self.backup_dir = None

    def __repr__(self):
        repr_str = ', '.join(f'{attr}={repr(getattr(self, attr))}' for attr in self.REPR_ATTRS)
        return f'{self.__class__.__name__}({repr_str})'

    def get_subpath(self, file):
        """Get subpath of a file related to top_dir.

        Also convert "\" to "/", which makes it useful for showing a file in
            issue safely.
        """
        return os.path.relpath(file, self.root).replace('\\', '/')

    def get_tree_file(self, name, index=0):
        return os.path.join(self.tree_dir, f'{name}{index or ""}.js')

    def iter_tree_files(self, name):
        i = 0
        while True:
            file = self.get_tree_file(name, i)
            if not os.path.exists(file):
                break
            yield file
            i += 1

    def iter_meta_files(self):
        yield from self.iter_tree_files('meta')

    def iter_toc_files(self):
        yield from self.iter_tree_files('toc')

    def iter_fulltext_files(self):
        yield from self.iter_tree_files('fulltext')

    def load_tree_file(self, file):
        """Load a tree file.

        Raises:
            OSError: failed to open or read (unlikely to happen as this is
                usually called via iter_tree_files() and file existence has
                been checked in prior)
            TreeFileMalformedError: file malformed
        """
        with open(file, encoding='UTF-8') as fh:
            text = fh.read()

        m = self.REGEX_TREE_FILE_WRAPPER.search(text)

        if not m:
            raise TreeFileMalformedWrappingError(f'Malformed tree file wrapping', filename=file)

        try:
            return json.loads(m.group(1))
        except json.decoder.JSONDecodeError as exc:
            raise TreeFileMalformedJsonError(f'Malformed tree file: {exc}', filename=file) from exc

    def load_tree_files(self, name):
        data = {}
        for file in self.iter_tree_files(name):
            data.update(self.load_tree_file(file))

        # remove top-level None values to allow quick clear by appending file
        # e.g. add meta1.js with {'id1': None} to quickly delete 'id1' in meta.js
        for k in list(data):
            if data[k] is None:
                del data[k]

        return data

    def load_meta_files(self, refresh=False):
        if refresh or self.meta is None:
            self.meta = self.load_tree_files('meta')

    def load_toc_files(self, refresh=False):
        if refresh or self.toc is None:
            self.toc = self.load_tree_files('toc')

    def load_fulltext_files(self, refresh=False):
        if refresh or self.fulltext is None:
            self.fulltext = self.load_tree_files('fulltext')

    def save_tree_file(self, name, index, data):
        """Save a tree file.

        Raises:
            OSError: failed to write
        """
        file = self.get_tree_file(name, index)
        self.backup(file)
        with open(file, 'w', encoding='UTF-8', newline='\n') as fh:
            fh.write(data)

    def save_meta_file(self, i, data):
        self.save_tree_file('meta', i, f"""/**
 * Feel free to edit this file, but keep data code valid JSON format.
 */
scrapbook.meta({json.dumps(data, ensure_ascii=False, indent=2)})""")

    def save_meta_files(self):
        """Save to tree/meta#.js

        A javascript string >= 256 MiB (UTF-16 chars) causes an error
        in the browser. Split each js file at around 256 K items to
        prevent the issue. (An item is mostly < 512 bytes)
        """
        os.makedirs(os.path.join(self.tree_dir), exist_ok=True)
        i = 0
        size = 1
        meta = {}
        for id in list(self.meta):
            if self.meta[id] is None:
                del self.meta[id]
                continue
            meta[id] = self.meta[id]
            size += 1
            if size >= self.SAVE_META_THRESHOLD:
                self.save_meta_file(i, meta)
                i += 1
                size = 0
                meta = {}

        if size:
            self.save_meta_file(i, meta)
            i += 1

        # remove unused tree/meta#.js
        while True:
            file = self.get_tree_file('meta', i)
            try:
                os.remove(file)
            except FileNotFoundError:
                break
            i += 1

    def save_toc_file(self, i, data):
        self.save_tree_file('toc', i, f"""/**
 * Feel free to edit this file, but keep data code valid JSON format.
 */
scrapbook.toc({json.dumps(data, ensure_ascii=False, indent=2)})""")

    def save_toc_files(self):
        """Save to tree/toc#.js

        A javascript string >= 256 MiB (UTF-16 chars) causes an error
        in the browser. Split each js file at around 4 M entries to
        prevent the issue. (An entry is mostly < 32 bytes)
        """
        os.makedirs(os.path.join(self.tree_dir), exist_ok=True)
        i = 0
        size = 1
        toc = {}
        for id in list(self.toc):
            if self.toc[id] is None:
                del self.toc[id]
                continue
            toc[id] = self.toc[id]
            size += 1 + len(toc[id])
            if size >= self.SAVE_TOC_THRESHOLD:
                self.save_toc_file(i, toc)
                i += 1
                size = 0
                toc = {}

        if size:
            self.save_toc_file(i, toc)
            i += 1

        # remove unused tree/toc#.js
        while True:
            file = self.get_tree_file('toc', i)
            try:
                os.remove(file)
            except FileNotFoundError:
                break
            i += 1

    def save_fulltext_file(self, i, data):
        self.save_tree_file('fulltext', i, f"""/**
 * This file is generated by WebScrapBook and is not intended to be edited.
 */
scrapbook.fulltext({json.dumps(data, ensure_ascii=False, indent=1)})""")

    def save_fulltext_files(self):
        """Save to tree/fulltext#.js

        A javascript string >= 256 MiB (UTF-16 chars) causes an error
        in the browser. Split each js file at at around 128 MiB to
        prevent the issue.
        """
        os.makedirs(os.path.join(self.tree_dir), exist_ok=True)
        i = 0
        size = 1
        fulltext = {}
        for id in list(self.fulltext):
            if self.fulltext[id] is None:
                del self.fulltext[id]
                continue
            fulltext[id] = self.fulltext[id]
            for path in fulltext[id]:
                size += len(fulltext[id][path]['content'])
            if size >= self.SAVE_FULLTEXT_THRESHOLD:
                self.save_fulltext_file(i, fulltext)
                i += 1
                size = 0
                fulltext = {}

        if size:
            self.save_fulltext_file(i, fulltext)
            i += 1

        # remove unused tree/fulltext#.js
        while True:
            file = self.get_tree_file('fulltext', i)
            try:
                os.remove(file)
            except FileNotFoundError:
                break
            i += 1

    def init_backup(self, ts=True):
        """Setup a backup dir for following backups until next set.

        Args:
            ts: a webscrapbook ID as timestamp. True to generate one from
            current time. False to disable backup.
        """
        if ts is False:
            self.backup_dir = None
            return

        if ts is True:
            ts = util.datetime_to_id()

        self.backup_dir = os.path.join(self.host.backup_dir, ts)

    def backup(self, file, **kwargs):
        """Create a backup for a file or directory if self.backup_dir is set.
        """
        if not self.backup_dir:
            return

        self.host.backup(file, self.backup_dir, **kwargs)

    def get_lock(self, name, *args, **kwargs):
        return self.host.get_lock(f'book-{self.id}-{name}', *args, **kwargs)

    def get_tree_lock(self, *args, **kwargs):
        return self.get_lock('tree', *args, **kwargs)

    def get_index_paths(self, index):
        if util.is_maff(index):
            pages = util.get_maff_pages(os.path.join(self.data_dir, index))
            return [p.indexfilename for p in pages]

        if util.is_htz(index):
            return ['index.html']

        return [os.path.basename(index)]

    def get_tree_from_index_file(self, file):
        try:
            if util.is_html(file):
                with open(file, 'rb') as fh:
                    return util.load_tree(fh)

            if util.is_htz(file):
                with zipfile.ZipFile(file) as zh:
                    with zh.open('index.html') as fh:
                        fh = zip_stream(fh)
                        return util.load_tree(fh)

            if util.is_maff(file):
                info = next(iter(util.get_maff_pages(file)), None)
                if not info:
                    return None

                with zipfile.ZipFile(file) as zh:
                    with zh.open(info.indexfilename) as fh:
                        fh = zip_stream(fh)
                        return util.load_tree(fh)
        except (OSError, zipfile.BadZipFile, KeyError):
            return None

        return None

    def get_icon_file(self, item):
        """Get favicon file path of an item.

        Returns:
            str file path of the favicon, or None if not determinable
        """
        icon = item.get('icon', '')

        if not icon:
            return None

        u = urlsplit(icon)

        if u.scheme:
            return None

        if u.netloc:
            return None

        if not u.path:
            return None

        if u.path.startswith('/'):
            return None

        index = item.get('index', '')
        return os.path.normpath(os.path.join(self.data_dir, os.path.dirname(index), unquote(u.path)))

    def load_note_file(self, file):
        with open(file, encoding='UTF-8') as fh:
            content = fh.read()

        return self.REGEX_ITEM_NOTE.sub(r'\1', content)

    def save_note_file(self, file, content):
        data = util.format_string(self.ITEM_NOTE_FORMATTER, {
            'NOTE_CONTENT': content,
            })
        # enforce LF to prevent bad parsing for legacy ScrapBook
        with open(file, 'w', encoding='UTF-8', newline='\n') as fh:
            fh.write(data)
